# coding: utf-8

"""

    Image Processing
    ~~~~~~~~~~~~~~~~

"""
import os
import json
import requests
import numpy as np
import pandas as pd
from PIL import Image
from PIL import ImageStat
from scipy.misc import imread
from collections import defaultdict
from os.path import join as os_join
from six.moves.urllib.parse import urljoin
from os.path import basename as os_basename

# General tools
from biovida.support_tools.support_tools import (tqdm,
                                                 items_null,
                                                 data_frame_col_drop,
                                                 list_to_bulletpoints)

# Tools form the image subpackage
from biovida.images._image_tools import load_and_scale_images

# Open-i Support Tools
from biovida.images._interface_support.openi.openi_support_tools import (nonessential_openi_columns,
                                                                         grayscale_openi_modalities)

# Models
from biovida.images.models.border_detection import border_detection
from biovida.images.models.template_matching import robust_match_template
from biovida.images.models.image_classification import ImageClassificationCNN

# Suppress Pandas' SettingWithCopyWarning
pd.options.mode.chained_assignment = None


class OpeniImageProcessing(object):
    """

    This class is designed to allow easy analysis and cleaning of cached Open-i image data.

    :param instance: an instance of the ``biovida.images.openi_interface.OpenInterface()`` class.
    :type instance: ``OpenInterface Class``
    :param db_to_extract: ``records_db`` or``cache_records_db``. Defaults to 'records_db'.
    :type db_to_extract: ``str``
    :param model_location: the location of the model for Convnet.
                           If `None`, the default model will be used. Defaults to ``None``.
    :type model_location: ``str``
    :param download_override: If ``True``, download a new copy of the 'visual_image_problems_model' weights (and
                              associated resources) regardless of whether or not files with these names are
                              already cached. Defaults to ``False``.
    :type download_override: ``bool``
    :param verbose: if ``True``, print additional details. Defaults to ``False``.
    :type verbose: ``bool``

    :var image_dataframe: this is the dataframe that was passed when instantiating the class and
                          contains a cache of all analyses run as new columns.
    """

    def _obtain_model_resources(self, download_override):
        """
        
        Obtain the 'visual_image_problems_model' 
        
        :param download_override: If ``True``, download the 'visual_image_problems_model' weights (and
                                  associated resources) regardless of whether or not these files are already cached.
        :type download_override: ``bool``
        """
        resources_path = os_join(self.instance._created_image_dirs['aux'], 'image_processing_resources')
        if not os.path.isdir(resources_path):
            os.makedirs(resources_path)

        base_url = 'https://raw.githubusercontent.com/TariqAHassan/BioVida/master/data/model_resources/'

        required_resources = ["trained_open_i_modality_types.json",
                              "visual_image_problems_model.h5",
                              "visual_image_problems_model_support.p"]

        def download(url, file_path):
            response = requests.get(url)
            with open(file_path, 'wb') as file:
                file.write(response.content)

        for resource in required_resources:
            file_path = os_join(resources_path, resource)
            if not os.path.isfile(file_path) or download_override:
                if self._verbose:
                    print("Downloading '{0}'... ".format(resource))
                download(url=urljoin(base_url, resource), file_path=file_path)

        self._model_path = os_join(resources_path, "visual_image_problems_model.h5")

        # Load 'trained_open_i_modality_types.json' into memory
        with open(os_join(resources_path, 'trained_open_i_modality_types.json')) as json_data:
            self.trained_open_i_modality_types = json.load(json_data)

    @staticmethod
    def _extract_db(instance, db_to_extract):
        """

        Extracts a database from the `instance` parameter of OpeniImageProcessing().

        :param db_to_extract: see ``OpeniImageProcessing()``.
        :type db_to_extract: ``str``
        :param instance: see ``OpeniImageProcessing()``.
        :type instance: ``OpenInterface Class``
        :return: extract database
        :rtype: ``Pandas DataFrame``
        """
        if db_to_extract not in ('records_db', 'cache_records_db'):
            raise ValueError("`db_to_extract` must be one of: 'records_db', 'cache_records_db'.")

        extract = getattr(instance, db_to_extract)

        if isinstance(extract, pd.DataFrame):
            image_dataframe = extract.copy(deep=True)
            return image_dataframe
        else:
            raise TypeError("The '{0}' of `instance` must be of "
                            "type DataFrame, not: '{1}'.".format(db_to_extract, type(extract).__name__))

    def __init__(self,
                 instance,
                 db_to_extract='records_db',
                 model_location=None,
                 download_override=False,
                 verbose=True):
        self._verbose = verbose
        self.db_to_extract = db_to_extract
        self._cache_path = getattr(instance, '_cache_path')
        self.known_image_problems = ('arrows', 'asterisks', 'grids')

        if "OpeniInterface" != type(instance).__name__:
            raise ValueError("`instance` must be a `OpeniInterface` instance.")
        self.instance = instance

        # Extract the records_db/cache_records_db database
        self.image_dataframe = self._extract_db(instance, db_to_extract)
        self.image_dataframe_cleaned = None

        if 'cached_images_path' not in self.image_dataframe.columns:
            raise KeyError("No 'cached_images_path' column in '{0}'.".format(db_to_extract))

        # Extract path to the MedPix Logo
        self._medpix_path = instance._created_image_dirs['medpix_logo']

        # Load the CNN
        self._ircnn = ImageClassificationCNN()
        self.trained_open_i_modality_types = None
        self._model_path = None

        self._obtain_model_resources(download_override=download_override)

        # Load the model weights and architecture.
        if model_location is None:
            self._ircnn.load(self._model_path, override_existing=True)
        elif not isinstance(model_location, str):
            raise ValueError("`model_location` must either be a string or `None`.")
        elif os.path.isfile(model_location):
            self._ircnn.load(self._model_path, override_existing=True)
        else:
            raise FileNotFoundError("'{0}' could not be located.".format(str(model_location)))

        # Load the visual image problems the model can detect
        self.model_classes = list(self._ircnn.data_classes.keys())

        # Container for images represented as `ndarrays`
        self._ndarrays_images = None

    @property
    def image_dataframe_short(self):
        """Return `image_dataframe` with nonessential columns removed."""
        return data_frame_col_drop(self.image_dataframe, nonessential_openi_columns, 'image_dataframe')

    def _pil_load(self, image_paths, convert_to_rgb, status):
        """

        Load an image from a list of paths using the ``PIL`` Library.

        :param image_paths: images paths to load.
        :type image_paths: ``list``, ``tuple`` or ``Pandas Series``
        :param convert_to_rgb: if True, convert the image to RGB.
        :type convert_to_rgb: ``bool``
        :param status: display status bar. Defaults to True.
        :type status: ``bool``
        :return: a list of a. PIL images or b. PIL images in tuples of the form (PIL Image, image path).
        :rtype: ``list``
        """
        def conversion(image):
            return (Image.open(image).convert('RGB'), image) if convert_to_rgb else (Image.open(image), image)
        return [conversion(i) for i in tqdm(image_paths, desc="Loading Images", disable=not status)]

    def _ndarray_extract(self, zip_with_column=None, reload_override=False):
        """

        Loads images as `ndarrays` and flattens them.

        :param zip_with_column: a column from the `image_dataframe` to zip with the images. Defaults to None.
        :type zip_with_column: ``str``
        :param reload_override: if True, reload the images from disk. Defaults to False.
        :type reload_override: ``bool``
        :return: images as 2D ndarrays.
        :rtype: ``zip`` or ``list of ndarrays``
        """
        if self._ndarrays_images is None or reload_override:
            self._ndarrays_images = [imread(i, flatten=True) for i in self.image_dataframe['cached_images_path']]

        if zip_with_column is not None:
            return list(zip(*[self._ndarrays_images, self.image_dataframe[zip_with_column]]))
        else:
            return self._ndarrays_images

    @staticmethod
    def _grayscale_image(image_path):
        """

        Computes whether or not an image is grayscale.
        See the `grayscale_analysis()` method for caveats.

        :param image_path: path to an image.
        :type image_path:``str``
        :return: ``True`` if grayscale, else ``False``.
        :rtype: ``bool``
        """
        # See: http://stackoverflow.com/q/23660929/4898004
        if image_path is None or items_null(image_path):
            return np.NaN
        image = Image.open(image_path)  # ToDo: find way to call only once inside this class (similar to _ndarray_extract())
        stat = ImageStat.Stat(image.convert("RGB"))
        return np.mean(stat.sum) == stat.sum[0]

    def grayscale_analysis(self, new_analysis=False, status=True):
        """

        Analyze the images to determine whether or not they are grayscale
        (uses the PIL image library).

        Note:
              - this tool is very conservative (very small amounts of 'color' will yield `False`).
              - the exception to the above rule is the *very rare* case of an image which even split
                of red, green and blue.

        :param new_analysis: rerun the analysis if it has already been computed. Defaults to ``False``.
        :type new_analysis: ``bool``
        :param status: display status bar. Defaults to ``True``.
        :type status: ``bool``
        """
        if 'grayscale' not in self.image_dataframe.columns or new_analysis:
            column = self.image_dataframe['cached_images_path']
            grayscale = [self._grayscale_image(i) for i in tqdm(column, desc='Grayscale Analysis', disable=not status)]
            self.image_dataframe['grayscale'] = grayscale

    @staticmethod
    def _logo_analysis_out(analysis_results, output_params):
        """

        Decides the output for the ``logo_analysis`` function.
        If the bonding box is in an improbable location, NaN is returned.
        Otherwise, the bonding box, or some portion of it (i.e., the lower left) will be returned.

        :param analysis_results: the output of ``biovida.images.models.template_matching.robust_match_template()``.
        :type analysis_results: ``dict``
        :return: the output requested by the ``logo_analysis()`` method.
        :rtype: ``NaN``, ``dict`` or ``tuple``
        """
        # Unpack ``output_params``
        match_quality_threshold, x_greater_check, y_greater_check = output_params

        # Unpack ``analysis_results``
        bounding_box = analysis_results['bounding_box']
        match_quality = analysis_results['match_quality']
        base_image_shape = analysis_results['base_image_shape']

        # Check match quality.
        if bounding_box is None or match_quality < match_quality_threshold:
            return np.NaN

        # Check the box is in the top right (as defined by ``x_greater_check`` and ``y_greater_check``).
        if bounding_box['bottom_left'][0] < (base_image_shape[0] * x_greater_check) or \
                        bounding_box['bottom_left'][1] > (base_image_shape[1] * y_greater_check):
            return np.NaN

        return bounding_box

    def _logo_processor(self, robust_match_template_wrapper, output_params, status):
        """

        Performs the actual analysis for ``logo_analysis()``, searching for the
        MedPix logo in the images.

        :param robust_match_template_wrapper: wrapper generated inside of ``logo_analysis()``
        :type robust_match_template_wrapper: ``func``
        :param output_params: tuple of the form:

                        ``(match_quality_threshold, xy_position_threshold[0], xy_position_threshold[1])``

        :type output_params: ``tuple``
        :param status: display status bar. Defaults to ``True``.
        :type status: ``bool``
        :return: a list of dictionaries or tuples specifying all, or part of (i.e., lower left) the bonding box
                 for the pattern in the base image.
        :rtype: ``list of dicts`` or ``list of tuples``
        """
        results = list()

        # Use images in the dataframe represented as ndarrays, along with
        # the journal title (to check for their source being medpix).
        to_analyze = self._ndarray_extract(zip_with_column='journal_title')

        for image, journal in tqdm(to_analyze, desc='Logo Analysis', disable=not status):
            if 'medpix' not in str(journal).lower():
                results.append(np.NaN)
            else:
                analysis_results = robust_match_template_wrapper(image)
                current = self._logo_analysis_out(analysis_results, output_params)
                results.append(current)

        return results

    def logo_analysis(self,
                      match_quality_threshold=0.25,
                      xy_position_threshold=(1 / 3.0, 1 / 2.5),
                      base_resizes=(0.5, 2.5, 0.1),
                      end_search_threshold=0.875,
                      base_image_cropping=(0.15, 0.5),
                      new_analysis=False,
                      status=True):
        """

        Search for the MedPix Logo. If located, with match quality above match_quality_threshold,
        populate the the 'medpix_logo_bounding_box' of ``image_dataframe`` with its bounding box.

        :param match_quality_threshold: the minimum match quality required to accept the match.
                                        See: ``skimage.feature.match_template()`` for more information.
        :type match_quality_threshold: ``float``
        :param xy_position_threshold: tuple of the form: (x_greater_check, y_greater_check).
                                      For instance the default (``(1/3.0, 1/2.5)``) requires that the
                                      x position of the logo is greater than 1/3 of the image's width
                                      and less than 1/2.5 of the image's height.
        :type xy_position_threshold: ``tuple``
        :param base_resizes: See: ``biovida.images.models.template_matching.robust_match_template()``.
        :type base_resizes: ``tuple``
        :param end_search_threshold: See: ``biovida.images.models.template_matching.robust_match_template()``.
        :type end_search_threshold: ``float``
        :param base_image_cropping: See: ``biovida.images.models.template_matching.robust_match_template()``
        :type base_image_cropping: ``tuple``
        :param new_analysis: rerun the analysis if it has already been computed.
        :type new_analysis: ``bool``
        :param status: display status bar. Defaults to ``True``.
        :type status: ``bool``
        """
        # Note: this method wraps ``biovida.images.models.template_matching.robust_match_template()``.
        if 'medpix_logo_bounding_box' in self.image_dataframe.columns and not new_analysis:
            return None

        # Package Params
        output_params = (match_quality_threshold, xy_position_threshold[0], xy_position_threshold[1])

        # Load the Pattern. ToDo: Allow for non MedPix logos logos.
        medpix_template_image = imread(self._medpix_path, flatten=True)

        def robust_match_template_wrapper(image):
            return robust_match_template(pattern_image=medpix_template_image,
                                         base_image=image,
                                         base_resizes=base_resizes,
                                         end_search_threshold=end_search_threshold,
                                         base_image_cropping=base_image_cropping)

        # Run the algorithm searching for the medpix logo in the base image
        self.image_dataframe['medpix_logo_bounding_box'] = self._logo_processor(robust_match_template_wrapper,
                                                                                output_params=output_params,
                                                                                status=status)

    def border_analysis(self,
                        signal_strength_threshold=0.25,
                        min_border_separation=0.15,
                        lower_bar_search_space=0.9,
                        new_analysis=False,
                        status=True):
        """

        Wrapper for ``biovida.images.models.border_detection.border_detection()``.

        :param signal_strength_threshold: see ``biovida.images.models.border_detection()``.
        :type signal_strength_threshold: ``float``
        :param min_border_separation: see ``biovida.images.models.border_detection()``.
        :type min_border_separation: ``float``
        :param lower_bar_search_space: see ``biovida.images.models.border_detection()``.
        :type lower_bar_search_space: ``float``
        :param new_analysis: rerun the analysis if it has already been computed. Defaults to ``False``.
        :type new_analysis: ``bool``
        :param status: display status bar. Defaults to ``True``.
        :type status: ``bool``
        """
        if all(x in self.image_dataframe.columns for x in ['hbar', 'hborder', 'vborder']) and not new_analysis:
            return None

        def ba_func(image):
            return border_detection(image,
                                    signal_strength_threshold=signal_strength_threshold,
                                    min_border_separation=min_border_separation,
                                    lower_bar_search_space=lower_bar_search_space,
                                    report_signal_strength=False,
                                    rescale_input_ndarray=True)

        to_analyze = self._ndarray_extract()

        # Run the analysis
        border_analysis = [ba_func(i) for i in tqdm(to_analyze, desc='Border Analysis', disable=not status)]

        # Convert to a dataframe
        ba_df = pd.DataFrame(border_analysis).fillna(np.NaN)

        # Update datafame
        self.image_dataframe['hbar'] = ba_df['hbar']
        self.image_dataframe['hborder'] = ba_df['hborder']
        self.image_dataframe['vborder'] = ba_df['vborder']

    @staticmethod
    def _h_crop_top_decision(x):
        """

        Choose lowest horizontal cropping point.
        Solves: upper 'hborder' vs 'medpix_logo_bounding_box'
        (dictionary key = 'bottom_left').

        :param x: data passed through Pandas ``DataFrame.apply()`` method.
        :type x: ``Pandas Object``
        :return: the lowest crop location.
        :rtype: ``int`` or ``float``
        """
        # Note: hborder = [top, lower]; medpix_logo_bounding_box['bottom_left'] = [x, y].
        # That is, `hborder[0]` must be compared with `medpix_logo_bounding_box['bottom_left'][1]`.
        crop_candidates = list()
        if isinstance(x['hborder'], (list, tuple)):
            crop_candidates.append(x['hborder'][0])
        if isinstance(x['medpix_logo_bounding_box'], dict):
            lower_left = x['medpix_logo_bounding_box'].get('bottom_left', None)
            if isinstance(lower_left, (list, tuple)):
                crop_candidates.append(lower_left[1])

        return max(crop_candidates) if len(crop_candidates) else np.NaN

    @staticmethod
    def _h_crop_lower_decision(x):
        """

        Chose the highest cropping point for the image's bottom.
        Solves: lower 'hborder' vs 'hbar'.

        :param x: data passed through Pandas ``DataFrame.apply()`` method.
        :type x: ``Pandas Object``
        :return: the highest crop location.
        :rtype: ``int`` or ``float``
        """
        # Note: hborder = [top, lower]; hbar = int.
        lhborder = [x['hborder'][1]] if not items_null(x['hborder']) else []
        hbar = [x['hbar']] if not items_null(x['hbar']) else []
        crop_candidates = lhborder + hbar
        return min(crop_candidates) if len(crop_candidates) else np.NaN

    def crop_decision(self, new_analysis=False):
        """
        
        Decide where to crop the images, if at all.

        :param new_analysis: rerun the analysis if it has already been computed. Defaults to ``False``.
        :type new_analysis: ``bool``
        """
        if all(x in self.image_dataframe.columns for x in ['upper_crop', 'lower_crop']) and not new_analysis:
            return None

        for i in ('medpix_logo_bounding_box', 'hborder', 'hbar'):
            if i not in self.image_dataframe.columns:
                raise KeyError("The `image_dataframe` does not contain the\nfollowing required column: '{0}'.\n"
                               "Please execute the corresponding analysis method to generate it.".format(i))

        # Compute Crop location
        self.image_dataframe['upper_crop'] = self.image_dataframe.apply(self._h_crop_top_decision, axis=1)
        self.image_dataframe['lower_crop'] = self.image_dataframe.apply(self._h_crop_lower_decision, axis=1)

    @staticmethod
    def _apply_cropping(cached_images_path,
                        lower_crop,
                        upper_crop,
                        vborder,
                        return_as_array=True,
                        convert_to_rgb=True):
        """

        Applies cropping to a specific image.

        :param cached_images_path: path to the image
        :type cached_images_path: ``str``
        :param lower_crop: row of the column produced by the ``crop_decision()`` method.
        :type lower_crop: ``int`` (can be ``float`` if the column contains NaNs)
        :param upper_crop: row of the column produced by the ``crop_decision()`` method.
        :type upper_crop: ``int`` (can be ``float`` if the column contains NaNs)
        :param vborder: yield of the ``border_analysis()`` method
        :type vborder: ``int`` (can be ``float`` if the column contains NaNs)
        :param return_as_array: if True, convert the PIL object to an ``ndarray``. Defaults to True.
        :type return_as_array: ``bool``
        :return: the cropped image as either a PIL image or 2D ndarray.
        :rtype: ``PIL`` or ``2D ndarray``
        """
        # Load the image
        converted_image = Image.open(cached_images_path)

        # Horizontal Cropping
        if not items_null(lower_crop):
            w, h = converted_image.size
            converted_image = converted_image.crop((0, 0, w, int(lower_crop)))
        if not items_null(upper_crop):
            w, h = converted_image.size
            converted_image = converted_image.crop((0, int(upper_crop), w, h))

        # Vertical Cropping
        if not items_null(vborder):
            w, h = converted_image.size
            converted_image = converted_image.crop((int(vborder[0]), 0, int(vborder[1]), h))

        if convert_to_rgb:
            image_to_save = converted_image.convert("RGB")
        else:
            image_to_save = converted_image

        if return_as_array:
            return np.asarray(image_to_save)
        else:
            return image_to_save

    def _cropper(self, data_frame=None, return_as_array=True, convert_to_rgb=True, status=True):
        """

        Uses `_apply_cropping()` to apply cropping to images in a dataframe.

        :param data_frame: a dataframe with 'cached_images_path', 'lower_crop', 'upper_crop' and 'vborder' columns.
                          If ``None`` ``image_dataframe`` is used.
        :type data_frame: ``None`` or ``Pandas DataFrame``
        :param return_as_array: if True, convert the PIL object to an ``ndarray``. Defaults to True.
        :type return_as_array: ``bool``
        :param convert_to_rgb: if True, use the PIL library to convert the images to RGB. Defaults to False.
        :type convert_to_rgb: ``bool``
        :param status: display status bar. Defaults to True.
        :type status: ``bool``
        :return: cropped PIL images.
        :rtype: ``list``
        """
        if isinstance(data_frame, pd.DataFrame):
            df = data_frame
        else:
            df = self.image_dataframe

        all_cropped_images = list()
        for index, row in tqdm(df.iterrows(), desc='Cropping Images', disable=not status, total=len(df)):
            cropped_image = self._apply_cropping(cached_images_path=row['cached_images_path'],
                                                 lower_crop=row['lower_crop'],
                                                 upper_crop=row['upper_crop'],
                                                 vborder=row['vborder'],
                                                 return_as_array=return_as_array,
                                                 convert_to_rgb=convert_to_rgb)
            all_cropped_images.append(cropped_image)

        return all_cropped_images

    def visual_image_problems(self, limit_to_known_modalities=True, new_analysis=False, status=True):
        """

        This method is powered by a Convolutional Neural Network which
        computes probabilities for the presence of problematic image properties or types.

        Currently, the model can identify the follow problems:

        - arrows in images
        - images arrayed as grids

        :param limit_to_known_modalities: if ``True``, remove model predicts for image modalities
                                          the model has not explicitly been trained on. Defaults to ``True``.
        :type limit_to_known_modalities: ``bool``
        :param new_analysis: rerun the analysis if it has already been computed. Defaults to ``False``.
        :type new_analysis: ``bool``
        :param status: display status bar. Defaults to ``True``.
        :type status: ``bool``


        :Examples:

        >>> DataFrame['visual_image_problems']
        ...
        0 [('valid_image', 0.82292306), ('text', 0.13276383), ('arrows', 0.10139297), ('grids', 0.021935554)]
        1 [('valid_image', 0.76374823), ('arrows', 0.1085605), ('grids', 0.0024915827), ('text', 0.00037114936)]
        2 [('valid_image', 0.84319711), ('text', 0.10483728), ('arrows', 0.06458132), ('grids', 0.0125442)]
        3 [('valid_image', 0.84013706), ('arrows', 0.090836897), ('text', 0.055015128), ('grids', 0.0088913934)]

        The first value in the tuple represents the problem identified and second
        value represents its associated probability.
        """
        if 'visual_image_problems' in self.image_dataframe.columns and not new_analysis:
            return None

        cropped_images_for_analysis = self._cropper(return_as_array=True, status=status)

        transformed_images = load_and_scale_images(list_of_images=cropped_images_for_analysis,
                                                   image_size=self._ircnn.image_shape, status=status,
                                                   desc='Preparing Images')

        # Scan Images for Visual Problems with Neural Network
        self.image_dataframe['visual_image_problems'] = self._ircnn.predict(list_of_images=[transformed_images],
                                                                            status=status, desc='Scanning for Problems')

        if limit_to_known_modalities:  # ToDo: Temporary. Future: avoid passing through the model in the first place.
            for index, row in self.image_dataframe.iterrows():
                if row['image_modality_major'] not in self.trained_open_i_modality_types:
                    self.image_dataframe.set_value(index, 'image_modality_major', np.NaN)

    def auto_analysis(self, limit_to_known_modalities=True, new_analysis=False, status=True):
        """

        Automatically use the class methods to analyze the ``image_dataframe`` using default
        parameter values for class methods.

        :param limit_to_known_modalities: if ``True``, remove model predicts for image modalities
                                          the model has not explicitly been trained on. Defaults to ``True``.
        :type limit_to_known_modalities: ``bool``
        :param new_analysis: rerun the analysis if it has already been computed. Defaults to ``False``.
        :type new_analysis: ``bool``
        :param status: display status bar. Defaults to ``True``.
        :type status: ``bool``
        """
        # Run Analysis Battery with Default Parameter Values
        self.grayscale_analysis(new_analysis=new_analysis, status=status)
        self.logo_analysis(new_analysis=new_analysis, status=status)
        self.border_analysis(new_analysis=new_analysis, status=status)

        # Compute Crop location
        self.crop_decision(new_analysis=new_analysis)

        # Generate predictions
        self.visual_image_problems(limit_to_known_modalities=limit_to_known_modalities,
                                   new_analysis=new_analysis, status=status)

    @staticmethod
    def _invalid_image_tests(row, problems_to_ignore, valid_floor, image_problem_threshold=None):
        """

        Tests to determine if ``row`` references an image with properties and/or features
        likely to be problematic when training a model.

        :param row: as passed by ``pandas.DataFrame.apply(func, axis=1)``.
        :type row: ``Pandas Series``
        :param problems_to_ignore: see ``auto_decision()``.
        :type problems_to_ignore: ``None``, ``list`` or ``tuple``
        :param valid_floor: see ``auto_decision()``.
        :type valid_floor: ``float``
        :param image_problem_threshold: a scalar from 0 to 1 which specifies the threshold value required
                                        to cause the image to be marked as invalid.
                                        For instance, a threshold value of `0.5` would mean that any image
                                        which contains a image problem probability above `0.5` will be marked
                                        as invalid. 
                                        
                                        NOTE: Currently not in use.
                                        
        :type image_problem_threshold: ``float``
        :return: a list of the form ``[invalid image (boolean), reasons for decision if the former is True]``, wrapped
                 in a pandas series so it can be neatly split into two columns when called via. ``DataFrame.apply()``.
        :rtype: ``Pandas Series``
        """
        def image_problems_from_text_test(ipft):
            problem = False
            if not isinstance(problems_to_ignore, (list, tuple)) and \
                    isinstance(ipft, (list, tuple)) and len(ipft):
                problem = True
            elif (isinstance(problems_to_ignore, (list, tuple)) and
                  isinstance(ipft, (list, tuple)) and len([i for i in ipft if i not in problems_to_ignore])):
                problem = True
            return ['image_problems_from_text'] if problem else []

        def visual_image_problems_test(vip):
            if isinstance(problems_to_ignore, (list, tuple)):
                vip_ = [i for i in vip if i[0] not in problems_to_ignore]
            else:
                vip_ = vip

            problem = False
            if len(vip_) == 1:
                if vip_[0][0] == 'valid_image' and vip_[0][1] < valid_floor:
                    problem = True
            else:
                if vip_[0][0] != 'valid_image':
                    problem = True
                elif vip_[0][0] == 'valid_image' and vip_[0][1] < valid_floor:
                    problem = True
            return ['visual_image_problems'] if problem else []

        reasons = list()
        if row['grayscale'] is False and row['image_modality_major'] in grayscale_openi_modalities:
            reasons.append('grayscale')
        if isinstance(row['image_problems_from_text'], (list, tuple)):
            reasons += image_problems_from_text_test(ipft=row['image_problems_from_text'])
        if isinstance(row['visual_image_problems'], (list, tuple)):
            reasons += visual_image_problems_test(vip=row['visual_image_problems'])

        return pd.Series([len(reasons) > 0, tuple(sorted(reasons)) if len(reasons) else None])

    def auto_decision(self, valid_floor=0.8, problems_to_ignore=None):
        """

        Automatically generate 'invalid_image' column in the `image_dataframe`
        column by deciding whether or not images are valid using default parameter values for class methods.

        :param valid_floor: the smallest value needed for a 'valid_image' to be considered valid. Defaults to `0.8`.
        
                .. note::
                
                    For an image to be considered 'valid', the most likely categorization for the image must be
                    'valid_image' and the probability that the model assigns when placing it in this category
                    must be great than or equal to ``valid_floor``.
        
        :type valid_floor: ``float``
        :param problems_to_ignore: image problems to ignore. See ``INSTANCE.known_image_problems`` for valid values.
                                   Defaults to ``None``.
        :type problems_to_ignore: ``None``, ``list`` or ``tuple``
        """
        for i in ('grayscale', 'image_problems_from_text', 'visual_image_problems'):
            if i not in self.image_dataframe.columns:
                raise KeyError("`image_dataframe` does not contain a '{0}' column.".format(i))

        if isinstance(problems_to_ignore, (list, tuple)):
            for i in problems_to_ignore:
                if i not in self.known_image_problems or i == 'valid_image':
                    raise ValueError("`problems_to_ignore` may only contain the following:\n"
                                     "{0}".format(list_to_bulletpoints(self.known_image_problems)))
        elif problems_to_ignore is not None:
            raise ValueError("`problems_to_ignore` must be a `string`, `list` or `tuple`.")

        test_results = self.image_dataframe.apply(
            lambda r: self._invalid_image_tests(r, problems_to_ignore, valid_floor), axis=1)
        self.image_dataframe['invalid_image'] = test_results[0]
        self.image_dataframe['invalid_image_reasons'] = test_results[1].fillna(np.NaN)

    def auto(self, valid_floor=0.8, limit_to_known_modalities=True,
             problems_to_ignore=None, new_analysis=False, status=True):
        """

        Automatically carry out all aspects of image preprocessing (recommended).

        :param valid_floor: the smallest value needed for a 'valid_image' to be considered valid. Defaults to `0.8`.
        :type valid_floor: ``float``
        :param limit_to_known_modalities: if ``True``, remove model predicts for image modalities
                                          the model has not explicitly been trained on. Defaults to ``True``.
        :type limit_to_known_modalities: ``bool``
        :param problems_to_ignore: image problems to ignore. See ``INSTANCE.known_image_problems`` for valid values.
                                   Defaults to ``None``.
        :type problems_to_ignore: ``None``, ``list`` or ``tuple``
        :param new_analysis: rerun the analysis if it has already been computed. Defaults to ``False``.
        :type new_analysis: ``bool``
        :param status: display status bar. Defaults to ``True``.
        :type status: ``bool``
        :return: the `image_dataframe`, complete with the results of all possible analyses
                (using default parameter values).
        :rtype: ``Pandas DataFrame``
        """
        # Run Auto Analysis
        self.auto_analysis(limit_to_known_modalities=limit_to_known_modalities,
                           new_analysis=new_analysis, status=status)

        # Run Auto Decision
        self.auto_decision(problems_to_ignore=problems_to_ignore,
                           valid_floor=valid_floor)

        return self.image_dataframe

    def _save_method_error_checking(self, output_rule, action):
        """

        Check for error that would cause ``save()`` to fail.

        :param output_rule: see ``output()``
        :type output_rule: ``str`` or ``func``
        :param action: see ``output``
        :type action: ``str``
        """
        if action not in ('copy', 'ndarray'):
            raise ValueError("`action` must be either 'copy' or 'ndarray'.")
        if not isinstance(output_rule, str) and not callable(output_rule):
            raise TypeError("`output_rule` must be a string or function.")
        if 'invalid_image' not in self.image_dataframe.columns:
            raise KeyError("`image_dataframe` must contain a 'invalid_image' column which uses booleans to\n"
                           "indicate whether or not to include an entry in the cleaned dataset.\n"
                           "To automate this process, consider using the `auto_decision()` method.")

    def clean_image_dataframe(self, crop_images=True, convert_to_rgb=False, status=True):
        """

        Define a dataframe with rows of images found to be 'valid'.
        These 'valid' images are cleaned and stored as PIL images in a ``'cleaned_image'`` column.
        
        .. note::
            
            The DataFrame this method creates can be viewed with ``INSTANCE.image_dataframe_cleaned``.

        :param crop_images: Crop the images using analyses results from `border_analysis()` and
                            ``logo_analysis()``. Defaults to ``True``.
        :type crop_images: ``bool``
        :param convert_to_rgb: if ``True``, use the PIL library to convert the images to RGB. Defaults to ``False``.
        :type convert_to_rgb: ``bool``
        :param status: display status bar. Defaults to ``True``.
        :type status: ``bool``
        :return: ``self.image_dataframe`` where the 'invalid_image' column is ``True``, with the addition of
                  a 'cleaned_image' populated by PIL image to be saved to disk.
        :rtype: ``Pandas DataFrame``
        """
        if 'invalid_image' not in self.image_dataframe:
            raise KeyError("No `invalid_image` column in `image_dataframe`,\n"
                           "which is required to determine output.\n"
                           "Consider calling the ``auto()`` method.")

        image_dataframe_cleaned = self.image_dataframe[
            self.image_dataframe['invalid_image'] != True].reset_index(drop=True).copy(deep=True)

        if crop_images:
            image_dataframe_cleaned['cleaned_image'] = self._cropper(data_frame=image_dataframe_cleaned,
                                                                     return_as_array=False,
                                                                     convert_to_rgb=convert_to_rgb,
                                                                     status=status)
        else:
            image_dataframe_cleaned['cleaned_image'] = self._pil_load(image_dataframe_cleaned['cached_images_path'],
                                                                      convert_to_rgb=convert_to_rgb, status=status)

        self.image_dataframe_cleaned = image_dataframe_cleaned

    def output(self,
               output_rule,
               create_dirs=False,
               allow_overwrite=True,
               action='copy',
               status=True,
               **kwargs):
        """

        Save processed images to disk.

        :param output_rule:

            - if a ``str``: the directory to save the images.
            - if a ``function``: it must (1) accept a single parameter (argument) and (2) return system path(s)
              [see example below].

        :type output_rule: ``str`` or ``func``
        :param create_dirs: if ``True``, create directories returned by ``divvy_rule`` if they do not exist.
                            Defaults to ``False``.
        :type create_dirs: ``bool``
        :param allow_overwrite: if ``True`` allow existing images to be overwritten. Defaults to ``True``.
        :type allow_overwrite: ``bool``
        :param action: one of 'copy', 'ndarray'.
        :type action: ``str``
        :param status: display status bar. Defaults to ``True``.
        :type status: ``bool``

        :Example:

        >>> from biovida.images import OpeniInterface
        >>> from biovida.images import OpeniImageProcessing
        ...
        >>> opi = OpeniInterface()
        >>> opi.search(image_type='mri')
        >>> opi.pull()
        ...
        >>> ip = OpeniImageProcessing(opi)
        >>> ip.auto()
        
        Next, prune invalid images 
        
        >>> ip.clean_image_dataframe()

        A Simple Output Rule

        >>> ip.output('/your/path/here/images')

        A More Complex Output Rule

        >>> def my_save_rule(row):
        >>>     if isinstance(row['abstract'], str) and 'lung' in row['abstract']:
        >>>         return '/your/path/here/lung_images'
        >>>     elif isinstance(row['abstract'], str) and 'heart' in row['abstract']:
        >>>         return '/your/path/here/heart_images'
        ...
        >>> ip.save(my_save_rule)

        """
        # Limit to 'valid' images
        if not isinstance(self.image_dataframe_cleaned, pd.DataFrame):
            raise TypeError("`image_dataframe_cleaned` is not a DataFrame.\n"
                            "The `clean_image_dataframe()` method must be called before ``output()``.")

        self._save_method_error_checking(output_rule=output_rule, action=action)
        allow_write = kwargs.get('allow_write', True)
        ndarray_with_path = kwargs.get('ndarray_with_path', False)

        def save_rule_wrapper(row):
            if isinstance(output_rule, str):
                group = output_rule
            elif callable(output_rule):
                group = output_rule(row)
                if group is None:
                    return None
                if not isinstance(group, str):
                    raise TypeError("String Expected.\nThe function passed to `output_rule` "
                                    "(`{0}`)\nreturned an object of type "
                                    "'{1}'.".format(output_rule.__name__, type(group).__name__))
            if action == 'copy':
                if os.path.isdir(group):
                    return group
                elif create_dirs and allow_write:
                    os.makedirs(group)
                    return group
                elif allow_write:
                    raise NotADirectoryError("\nNo such directory:\n'{0}'\n"
                                             "Consider setting `create_dirs=True`.".format(group))
            elif action == 'ndarray':
                return group

        return_dict = defaultdict(list)
        for _, row in tqdm(self.image_dataframe_cleaned.iterrows(), desc='Rendering Images',
                           disable=not status, total=len(self.image_dataframe_cleaned)):
            save_target = save_rule_wrapper(row)
            if isinstance(save_target, str):
                full_save_path = os_join(save_target, os_basename(row['cached_images_path']))
                if action == 'copy':
                    if allow_overwrite and allow_write:
                        row['cleaned_image'].save(full_save_path)
                    elif not os.path.isfile(full_save_path) and allow_write:
                        row['cleaned_image'].save(full_save_path)
                    return_dict[os_basename(save_target)] += [full_save_path]
                elif action == 'ndarray':
                    if ndarray_with_path:
                        return_dict[os_basename(save_target)] += [(np.array(row['cleaned_image']), full_save_path)]
                    else:
                        return_dict[os_basename(save_target)] += [np.array(row['cleaned_image'])]

        if action == 'ndarray':
            return {k: np.array(v) if not ndarray_with_path else v
                    for k, v in return_dict.items()}
        elif action == 'copy':
            return dict(return_dict)



# coding: utf-8

"""

    Support Data 
    ~~~~~~~~~~~~

"""

# To generate `age_dict`:
#     import inflect
#     from collections import OrderedDict
#     p = inflect.engine()
#     d = {p.number_to_words(i): i for i in range(1, 106)}
#     pprint(OrderedDict(sorted(d.items(), key=lambda x: x[1])))
#     Replace all '('; ')'. Replace "'," with "':".
age_dict = {'one': 1,
            'two': 2,
            'three': 3,
            'four': 4,
            'five': 5,
            'six': 6,
            'seven': 7,
            'eight': 8,
            'nine': 9,
            'ten': 10,
            'eleven': 11,
            'twelve': 12,
            'thirteen': 13,
            'fourteen': 14,
            'fifteen': 15,
            'sixteen': 16,
            'seventeen': 17,
            'eighteen': 18,
            'nineteen': 19,
            'twenty': 20,
            'twenty-one': 21,
            'twenty-two': 22,
            'twenty-three': 23,
            'twenty-four': 24,
            'twenty-five': 25,
            'twenty-six': 26,
            'twenty-seven': 27,
            'twenty-eight': 28,
            'twenty-nine': 29,
            'thirty': 30,
            'thirty-one': 31,
            'thirty-two': 32,
            'thirty-three': 33,
            'thirty-four': 34,
            'thirty-five': 35,
            'thirty-six': 36,
            'thirty-seven': 37,
            'thirty-eight': 38,
            'thirty-nine': 39,
            'forty': 40,
            'forty-one': 41,
            'forty-two': 42,
            'forty-three': 43,
            'forty-four': 44,
            'forty-five': 45,
            'forty-six': 46,
            'forty-seven': 47,
            'forty-eight': 48,
            'forty-nine': 49,
            'fifty': 50,
            'fifty-one': 51,
            'fifty-two': 52,
            'fifty-three': 53,
            'fifty-four': 54,
            'fifty-five': 55,
            'fifty-six': 56,
            'fifty-seven': 57,
            'fifty-eight': 58,
            'fifty-nine': 59,
            'sixty': 60,
            'sixty-one': 61,
            'sixty-two': 62,
            'sixty-three': 63,
            'sixty-four': 64,
            'sixty-five': 65,
            'sixty-six': 66,
            'sixty-seven': 67,
            'sixty-eight': 68,
            'sixty-nine': 69,
            'seventy': 70,
            'seventy-one': 71,
            'seventy-two': 72,
            'seventy-three': 73,
            'seventy-four': 74,
            'seventy-five': 75,
            'seventy-six': 76,
            'seventy-seven': 77,
            'seventy-eight': 78,
            'seventy-nine': 79,
            'eighty': 80,
            'eighty-one': 81,
            'eighty-two': 82,
            'eighty-three': 83,
            'eighty-four': 84,
            'eighty-five': 85,
            'eighty-six': 86,
            'eighty-seven': 87,
            'eighty-eight': 88,
            'eighty-nine': 89,
            'ninety': 90,
            'ninety-one': 91,
            'ninety-two': 92,
            'ninety-three': 93,
            'ninety-four': 94,
            'ninety-five': 95,
            'ninety-six': 96,
            'ninety-seven': 97,
            'ninety-eight': 98,
            'ninety-nine': 99,
            'one hundred': 100,
            'one hundred and one': 101,
            'one hundred and two': 102,
            'one hundred and three': 103,
            'one hundred and four': 104,
            'one hundred and five': 105
}

ages_as_tuples_rev = tuple(sorted(age_dict.items(), key=lambda x: x[-1], reverse=True))



"""

    Model Training
    ~~~~~~~~~~~~~~

    WARNING:

    This script is configured to use
    THEANO as a computational back end.
    To use TensorFlow, make the following
    change:

    K.set_image_dim_ordering('th') --> TO --> K.set_image_dim_ordering('tf')

"""
import sys

training_data_path = "/Users/tariq/Documents/biovida_training_data/biovida_train_val_test"
biovida_path = "/Users/tariq/Google Drive/Programming Projects/BioVida"
sys.path.append(biovida_path)

# General Imports
from keras import backend as K
K.set_image_dim_ordering('th')

from biovida.images.models.image_classification import ImageClassificationCNN


models = ['default', 'squeezenet', '_alex_net']


# ------------------------------------------------------------------------------------------
# Image Classification
# ------------------------------------------------------------------------------------------


def select_model():
    print("The following models can be used:\n")
    for e, m in enumerate(models, start=1):
        print("{0}.".format(str(e)), "'{0}'".format(m.replace("_", " ")))

    model_to_use_no = -1
    while model_to_use_no < 1 or model_to_use_no > len(models):
        model_to_use_no = int(input("Please enter the number of "
                                    "the model you wish to use: "))

    model_to_use = models[model_to_use_no - 1]
    print("Selected Model:", model_to_use)
    return model_to_use


def _image_recognition_cnn_training(epochs, model_to_use, training_data_path, save_name):
    """

    Train the model.

    :param epochs: number of epochs
    :type epochs: ``int``
    :param training_data_path: path to the synthesized_data data
    :type training_data_path: ``str``
    :param save_name: the name of the weights to be saved
    :type save_name: ``str``
    """
    ircnn = ImageClassificationCNN(training_data_path)
    ircnn.convnet(model_to_use=model_to_use)

    ircnn.fit(epochs=epochs)
    ircnn.save(save_name, path='/Users/tariq/Desktop')


save_name = "sq2" #input("Please enter 'save name' for the output file: ")
epochs = 2 #int(input("Please enter the number of epochs: "))
# model_to_use = select_model()
model_to_use = 'default'

_image_recognition_cnn_training(epochs, model_to_use, training_data_path, save_name)





# coding: utf-8

"""

    Image Classification
    ~~~~~~~~~~~~~~~~~~~~

"""
import os
import pickle
from warnings import warn

from biovida.support_tools.support_tools import tqdm
from biovida.images._image_tools import load_and_scale_images

# The model in `/_resources` was trained using Theano.
# TensorFlow cannot be used to load these weights until
# this issue (https://github.com/fchollet/keras/pull/5898) is resolved.
os.environ['KERAS_BACKEND'] = 'theano'

from keras import callbacks
from keras import backend as K
from keras.preprocessing.image import ImageDataGenerator
from keras.models import Sequential, load_model, Model
from keras.layers import (Convolution2D,
                          MaxPooling2D,
                          ZeroPadding2D,
                          concatenate,
                          GlobalAveragePooling2D,
                          Activation,
                          Dropout,
                          Flatten,
                          Dense,
                          Input,
                          merge)
from keras.optimizers import RMSprop, SGD

# Problem: ValueError: Negative dimension size caused by subtracting 2 from 1
# Solution: replace "tf" with "th" in ~/.keras/keras.json.
# Note: `MaxPooling2D` has a `dim_ordering` param which can do the same thing.

# ToDo: it's not currently obvious if keras.preprocessing.ImageDataGenerator().flow_from_directory()
# performs the preprocessing described in https://arxiv.org/pdf/1409.1556.pdf


class ImageClassificationCNN(object):
    """

    Keras Convolutional Neural Networks Interface.

    :param data_path: path to the directory with the subdirectories entitled 'train' and 'validation'.
                      This directory *must* have this structure. Defaults to ``None`` (to be use when loading
                      pre-computed weights).
    :type data_path: ``str``
    :param image_shape: the (height, width) to rescale the images to. Elements must be ``ints``. Defaults to ``(150, 150)``.
    :type image_shape: ``tuple`` or ``list``.
    :param rescale: See: ``keras.preprocessing.image.ImageDataGenerator()``. Defaults to 1/255.
    :type rescale: ``float``
    :param shear_range: See: ``keras.preprocessing.image.ImageDataGenerator()``. Defaults to 0.1.
    :type shear_range: ``float``
    :param zoom_range: See: ``keras.preprocessing.image.ImageDataGenerator()``. Defaults to 0.35.
    :type zoom_range: ``float``
    :param horizontal_flip: See: ``keras.preprocessing.image.ImageDataGenerator()``. Defaults to ``True``.
    :type horizontal_flip: ``bool``
    :param vertical_flip: See: ``keras.preprocessing.image.ImageDataGenerator()``. Defaults to ``True``.
    :type vertical_flip: ``bool``
    :param batch_size: Samples to propagate through the model.
                       See: ``keras.preprocessing.ImageDataGenerator().flow_from_directory()``.
                       Defaults to 4.
    :type batch_size: ``int``
    """

    def __init__(self,
                 data_path=None,
                 image_shape=(224, 224),
                 rescale=1/255.0,
                 shear_range=0.05,
                 zoom_range=0.30,
                 horizontal_flip=True,
                 vertical_flip=False,
                 batch_size=1):
        self._data_path = data_path
        self.image_shape = image_shape
        self.rescale = rescale
        self._shear_range = shear_range
        self._zoom_range = zoom_range
        self._horizontal_flip = horizontal_flip
        self._vertical_flip = vertical_flip
        self._batch_size = batch_size

        # Define data location
        if self._data_path is not None:
            self._train_data_dir = os.path.join(self._data_path, "train")
            self._validation_data_dir = os.path.join(self._data_path, "validation")

        # Data Streams
        self._train_generator = None
        self._validation_generator = None

        # The model itself
        self.model = None
        self.data_classes = None

    def _train_gen(self):
        """

        Use of ``keras.preprocessing.image.ImageDataGenerator()`` to generate training stream.

        """
        # Train augmentation configuration
        train_datagen = ImageDataGenerator(rescale=self.rescale,
                                           shear_range=self._shear_range,
                                           zoom_range=self._zoom_range,
                                           vertical_flip=self._vertical_flip,
                                           horizontal_flip=self._horizontal_flip)

        # Indefinitely generate batches of augmented train image data
        self._train_generator = train_datagen.flow_from_directory(directory=self._train_data_dir,
                                                                  target_size=self.image_shape,
                                                                  class_mode='categorical',
                                                                  batch_size=self._batch_size)

    def _val_gen(self):
        """

        Use of ``keras.preprocessing.image.ImageDataGenerator()`` to generate validation stream.

        """
        # Test augmentation configuration
        validation_datagen = ImageDataGenerator(rescale=self.rescale)

        # This is a similar generator, for validation data
        self._validation_generator = validation_datagen.flow_from_directory(directory=self._validation_data_dir,
                                                                            target_size=self.image_shape,
                                                                            class_mode='categorical',
                                                                            batch_size=self._batch_size)

    def _data_stream(self):
        """

        Generate Data Streams using ``keras.preprocessing.ImageDataGenerator()``.

        :raises: ``ValueError`` if there are asymmetries between the 'train' and 'validation'
                 subdirectories in ``self._data_path``.
        """
        # Create Data Streams
        self._train_gen()
        self._val_gen()

        train_classes = self._train_generator.class_indices
        val_classes = self._validation_generator.class_indices

        def set_diff(a, b):
            """Returns `True` if there are things in b not in a, else `False`."""
            return len(set(b) - set(a)) > 0

        # Check for a mismatch of folders between 'train' and 'validation'.
        chk = [(train_classes, val_classes, "train", "validation"), (val_classes, train_classes, "validation", "train")]
        for (i, j, k, l) in chk:
            if set_diff(i, j):
                raise ValueError("the `{0}` folder is missing the following"
                                 " folders found in '{1}': {2}.".format(k, l, ", ".join(map(str, set(j) - set(i)))))

        self.data_classes = train_classes

    def _impose_vgg_image_reqs(self):
        """

        This method imposes the 224x224 image size requirement made by VGG 19.

        """
        self.image_shape = list(self.image_shape)
        if self.image_shape[0] != 224:
            warn("{0} is an invalid image height for vgg_19. Falling back 224.".format(self.image_shape[0]))
            self.image_shape[0] = 224
        if self.image_shape[1] != 224:
            warn("{0} is an invalid image width for vgg_19. Falling back to 224.".format(self.image_shape[1]))
            self.image_shape[1] = 224
        self.image_shape = tuple(self.image_shape)

    def default(self, classes, output_layer_activation):
        """

        A Convolutional Neural Network with two convolution layers and ~5 million parameters.

        Sources:

        - https://blog.keras.io/building-powerful-image-classification-models-using-very-little-data.html

        :param classes: number of neuron in the output layer (which equals the number of classes).
        :type classes: ``int``
        :param output_layer_activation: the activation function to use on the output layer. See: https://keras.io/activations/#available-activations. Defaults to 'sigmoid'.
        :type output_layer_activation: ``str``
        """
        self.model = Sequential()
        self.model.add(Convolution2D(32, (3, 3),
                                     input_shape=(3, self.image_shape[0], self.image_shape[1]),
                                     activation='relu'))
        self.model.add(MaxPooling2D(pool_size=(2, 2)))

        self.model.add(Convolution2D(32, (3, 3), activation='relu'))
        self.model.add(MaxPooling2D(pool_size=(2, 2)))

        self.model.add(Flatten())
        self.model.add(Dense(64, activation='relu'))
        self.model.add(Dropout(0.5))
        self.model.add(Dense(classes))
        self.model.add(Activation(output_layer_activation))

    def _alex_net(self, classes, output_layer_activation):
        """

        .. warning:
        
            This is *NOT* stable with Keras 2.

        Sources:
        
        1. https://papers.nips.cc/paper/4824-imagenet-classification-with-deep-convolutional-neural-networks.pdf

        2. https://github.com/heuritech/convnets-keras/blob/master/convnetskeras/convnets.py

        :param classes: number of neuron in the output layer (which equals the number of classes).
        :type classes: ``int``
        :param output_layer_activation: the activation function to use on the output layer. See: https://keras.io/activations/#available-activations. Defaults to 'sigmoid'.
        :type output_layer_activation: ``str``
        """
        # Copyright (c) 2016 Heuritech
        #
        # Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
        # documentation files (the "Software"), to deal in the Software without restriction, including without limitation
        # the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
        # and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
        #
        # The above copyright notice and this permission notice shall be included in all copies or substantial portions
        # of the Software.
        #
        # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
        # TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
        # THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
        # CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
        # DEALINGS IN THE SOFTWARE.
        #
        # Note: not stable with Keras 2

        try:
            from convnetskeras.customlayers import crosschannelnormalization, splittensor
        except ImportError:
            raise ImportError("This method requires the 'convnets-keras' library.\n"
                              "This library can be installed by running the following command from the command line:\n"
                              "$ pip install git+git://github.com/tariqahassan/convnets-keras@master")

        inputs = Input(shape=(3, self.image_shape[0], self.image_shape[1]))

        conv_1 = Convolution2D(96, (11, 11),
                               subsample=(4, 4),
                               activation='relu',
                               name='conv_1')(inputs)

        conv_2 = MaxPooling2D((3, 3), strides=(2, 2))(conv_1)
        conv_2 = crosschannelnormalization(name="convpool_1")(conv_2)
        conv_2 = ZeroPadding2D((2, 2))(conv_2)
        conv_2 = merge([Convolution2D(128, (5, 5), activation="relu", name='conv_2_' + str(i + 1))(
                               splittensor(ratio_split=2, id_split=i)(conv_2)
                           ) for i in range(2)], mode='concat', concat_axis=1, name="conv_2")

        conv_3 = MaxPooling2D((3, 3), strides=(2, 2))(conv_2)
        conv_3 = crosschannelnormalization()(conv_3)
        conv_3 = ZeroPadding2D((1, 1))(conv_3)
        conv_3 = Convolution2D(384, (3, 3), activation='relu', name='conv_3')(conv_3)

        conv_4 = ZeroPadding2D((1, 1))(conv_3)
        conv_4 = merge([Convolution2D(192, 3, 3, activation="relu", name='conv_4_' + str(i + 1))(
                               splittensor(ratio_split=2, id_split=i)(conv_4)
                           ) for i in range(2)], mode='concat', concat_axis=1, name="conv_4")

        conv_5 = ZeroPadding2D((1, 1))(conv_4)
        conv_5 = merge([Convolution2D(128, (3, 3), activation="relu", name='conv_5_' + str(i + 1))(
                               splittensor(ratio_split=2, id_split=i)(conv_5)
                           ) for i in range(2)], mode='concat', concat_axis=1, name="conv_5")
        dense_1 = MaxPooling2D((3, 3), strides=(2, 2), name="convpool_5")(conv_5)

        dense_1 = Flatten(name="flatten")(dense_1)
        dense_1 = Dense(4096, activation='relu', name='dense_1')(dense_1)
        dense_2 = Dropout(0.5)(dense_1)
        dense_2 = Dense(4096, activation='relu', name='dense_2')(dense_2)
        dense_3 = Dropout(0.5)(dense_2)
        dense_3 = Dense(classes, name='dense_3')(dense_3)
        prediction = Activation(output_layer_activation, name=output_layer_activation)(dense_3)

        self.model = Model(input=inputs, output=prediction)

    def squeezenet(self, classes, output_layer_activation):
        """

        Keras Implementation of the SqueezeNet Model.

        Sources:
        
        - Iandola, Forrest N., Song Han, Matthew W. Moskewicz, Khalid Ashraf, William J. Dally, and Kurt Keutzer. 
          'SqueezeNet: AlexNet-level accuracy with 50x fewer parameters and< 0.5 MB model size.' arXiv preprint 
          `arXiv:1602.07360 <https://arxiv.org/abs/1602.07360>`_ (2016). 

        - Implementation adapted from: `github.com/rcmalli/keras-squeezenet <https://github.com/rcmalli/keras-squeezenet>`_

        :param classes: number of neuron in the output layer (which equals the number of classes).
        :type classes: ``int``
        :param output_layer_activation: the activation function to use on the output layer. See: https://keras.io/activations/#available-activations. Defaults to 'sigmoid'.
        :type output_layer_activation: ``str``
        """
        # MIT License
        #
        # Copyright (c) 2016 Refikcanmalli
        #
        # Permission is hereby granted, free of charge, to any person obtaining a copy
        # of this software and associated documentation files (the "Software"), to deal
        # in the Software without restriction, including without limitation the rights
        # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        # copies of the Software, and to permit persons to whom the Software is
        # furnished to do so, subject to the following conditions:
        #
        # The above copyright notice and this permission notice shall be included in all
        # copies or substantial portions of the Software.
        #
        # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        # SOFTWARE.
        #
        # Note: the model has been slightly modified. Namely, it has been refactored and
        # simplified by removing use of ImageNet weights.

        sq1x1 = "squeeze1x1"
        exp1x1 = "expand1x1"
        exp3x3 = "expand3x3"
        relu = "relu_"

        def fire_module(model, fire_id, squeeze=16, expand=64):
            s_id = 'fire' + str(fire_id) + '/'

            if K.image_data_format() == 'channels_first':
                channel_axis = 1
            else:
                channel_axis = 3

            model = Convolution2D(squeeze, (1, 1), padding='valid', name=s_id + sq1x1)(model)
            model = Activation('relu', name=s_id + relu + sq1x1)(model)

            left = Convolution2D(expand, (1, 1), padding='valid', name=s_id + exp1x1)(model)
            left = Activation('relu', name=s_id + relu + exp1x1)(left)

            right = Convolution2D(expand, (3, 3), padding='same', name=s_id + exp3x3)(model)
            right = Activation('relu', name=s_id + relu + exp3x3)(right)

            model = concatenate([left, right], axis=channel_axis, name=s_id + 'concat')
            return model

        input_img = Input(shape=(3, self.image_shape[0], self.image_shape[1]))

        model = Convolution2D(64, (3, 3), strides=(2, 2), padding='valid', name='conv1')(input_img)
        model = Activation('relu', name='relu_conv1')(model)
        model = MaxPooling2D(pool_size=(3, 3), strides=(2, 2), name='pool1')(model)

        model = fire_module(model, fire_id=2, squeeze=16, expand=64)
        model = fire_module(model, fire_id=3, squeeze=16, expand=64)
        model = MaxPooling2D(pool_size=(3, 3), strides=(2, 2), name='pool3')(model)

        model = fire_module(model, fire_id=4, squeeze=32, expand=128)
        model = fire_module(model, fire_id=5, squeeze=32, expand=128)
        model = MaxPooling2D(pool_size=(3, 3), strides=(2, 2), name='pool5')(model)

        model = fire_module(model, fire_id=6, squeeze=48, expand=192)
        model = fire_module(model, fire_id=7, squeeze=48, expand=192)
        model = fire_module(model, fire_id=8, squeeze=64, expand=256)
        model = fire_module(model, fire_id=9, squeeze=64, expand=256)
        model = Dropout(0.5, name='drop9')(model)

        model = Convolution2D(classes, (1, 1), padding='valid', name='conv10')(model)
        model = Activation('relu', name='relu_conv10')(model)
        model = GlobalAveragePooling2D()(model)
        output = Activation(output_layer_activation, name='loss')(model)

        self.model = Model(input_img, output, name='squeezenet')

    def vgg_19(self, classes, output_layer_activation):
        """

        Keras Implementation of the VGG19 Model.

        Sources:
        -------

        - Very Deep Convolutional Networks for Large-Scale Image Recognition
          K. Simonyan, A. Zisserman. `arXiv:1409.1556 <https://arxiv.org/abs/1409.1556>`_.
           
        - The Keras Implementation is available `here <https://gist.github.com/baraldilorenzo/8d096f48a1be4a2d660d#file-vgg-19_keras-py>`_.

        :param classes: number of neuron in the output layer (which equals the number of classes).
        :type classes: ``int``
        :param output_layer_activation: the activation function to use on the output layer. See: https://keras.io/activations/#available-activations. Defaults to 'sigmoid'.
        :type output_layer_activation: ``str``
        """
        self.model = Sequential()

        self.model.add(ZeroPadding2D((1, 1), input_shape=(3, self.image_shape[0], self.image_shape[1])))
        self.model.add(Convolution2D(64, (3, 3), activation='relu'))
        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(64, (3, 3), activation='relu'))
        self.model.add(MaxPooling2D((2, 2), strides=(2, 2)))

        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(128, (3, 3), activation='relu'))
        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(128, (3, 3), activation='relu'))
        self.model.add(MaxPooling2D((2, 2), strides=(2, 2)))

        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(256, (3, 3), activation='relu'))
        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(256, (3, 3), activation='relu'))
        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(256, (3, 3), activation='relu'))
        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(256, (3, 3), activation='relu'))
        self.model.add(MaxPooling2D((2, 2), strides=(2, 2)))

        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(512, (3, 3), activation='relu'))
        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(512, (3, 3), activation='relu'))
        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(512, (3, 3), activation='relu'))
        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(512, (3, 3), activation='relu'))
        self.model.add(MaxPooling2D((2, 2), strides=(2, 2)))

        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(512, (3, 3), activation='relu'))
        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(512, (3, 3), activation='relu'))
        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(512, (3, 3), activation='relu'))
        self.model.add(ZeroPadding2D((1, 1)))
        self.model.add(Convolution2D(512, (3, 3), activation='relu'))
        self.model.add(MaxPooling2D((2, 2), strides=(2, 2)))

        self.model.add(Flatten())
        self.model.add(Dense(4096, activation='relu'))
        self.model.add(Dropout(0.5))
        self.model.add(Dense(4096, activation='relu'))
        self.model.add(Dropout(0.5))
        self.model.add(Dense(classes, activation=output_layer_activation))

    def convnet(self,
                model_to_use='default',
                loss='binary_crossentropy',
                optimizer='default',
                metrics=('accuracy',),
                output_layer_activation='sigmoid'):
        """

        Define and Compile the Image Recognition Convolutional Neural Network.

        :param model_to_use: one of: 'default', 'squeezenet' or 'vgg19'. Defaults to 'default'.

            - 'default': a relatively simple sequential model with two convolution layers.

            - 'squeezenet': SqueezeNet model.

            - 'vgg19': the VGG 19 model.

        :type model_to_use: ``str``
        :param loss: Loss function. Defaults to 'categorical_crossentropy'.
                     See: ``keras.models.Sequential()``.
        :type loss: ``str``
        :param optimizer: Optimizer name. Defaults to 'default', which will use RMSprop with learning rate = ``0.0001``.
                          See: ``keras.models.Sequential()``.
        :type optimizer: ``str`` or ``keras.optimizers``
        :param metrics: Metrics to evaluate. Defaults to ('accuracy',).
                        Note: if round braces are used, it MUST contain a comma (to make it a tuple).
                        See: ``keras.models.Sequential()``.
        :type metrics: ``tuple``
        :param output_layer_activation: the activation function to use on the output layer. See: https://keras.io/activations/#available-activations. Defaults to 'sigmoid'.
        :type output_layer_activation: ``str``
        """
        if model_to_use == 'vgg19':
            self._impose_vgg_image_reqs()

        if self._data_path is not None:
            self._data_stream()

        classes = len(self.data_classes.keys())

        try:
            model_method = getattr(self, model_to_use)
        except:
            raise AttributeError("No such model: '{0}'.".format(model_to_use))

        model_method(classes, output_layer_activation=output_layer_activation)

        # Define optimizer
        if optimizer == 'default':
            optimizer_to_pass = RMSprop(lr=0.000025, rho=0.9, epsilon=1e-08, decay=0.0)
        elif optimizer == 'squeezenet':
            optimizer_to_pass = SGD(lr=0.001, decay=0.0002, momentum=0.9, nesterov=True)
        elif optimizer == 'vgg19':
            optimizer_to_pass = SGD(lr=0.1, decay=1e-6, momentum=0.9, nesterov=True)
        else:
            optimizer_to_pass = optimizer

        self.model.compile(loss=loss, optimizer=optimizer_to_pass, metrics=list(metrics))

    def _model_existence_check(self, first_format, second_format, additional=''):
        """

        Raises appropriate AttributeError based on content in which an undefined model was encountered.

        :param first_format: action the model
        :type first_format: ``str``
        :param second_format: name of an ``ImageClassificationCNN`` method.
        :type second_format: ``str``
        :raises: ``AttributeError`` composed from `first_format` and `second_format`.
        """
        if self.model is None:
            raise AttributeError("The model cannot be {0} until `ImageClassificationCNN().{1}()` "
                                 "has been called.{2}".format(first_format, second_format, additional))

    def fit(self, epochs=10, min_delta=0.1, patience=3):
        """

        Fit the model to the training data and run a validation.

        :param epochs: number of epochs. See: ``keras.models.Sequential()``. Defaults to 10.
        :type epochs: ``int``
        :param min_delta: see ``keras.callbacks.EarlyStopping()``.
        :type min_delta: ``float``
        :param patience: see ``keras.callbacks.EarlyStopping()``.
        :type patience: ``int``
        :raises: ``AttributeError`` if ``ImageClassificationCNN().convnet()`` is yet to be called.
        """
        self._model_existence_check("fit and validated", "convnet")

        if not isinstance(epochs, int):
            raise ValueError("`epochs` must be an integer.")

        early_stop = callbacks.EarlyStopping(monitor='val_loss', min_delta=min_delta, patience=patience, verbose=1)

        self.model.fit_generator(generator=self._train_generator,
                                 steps_per_epoch=self._train_generator.samples,
                                 epochs=epochs,
                                 validation_data=self._validation_generator,
                                 validation_steps=self._validation_generator.samples,
                                 callbacks=[early_stop])

    def _support_save_data(self, save_name, save_path):
        """

        Supporting Data For the Model.

        :param save_name: see ``save()``
        :type save_name: ``str``
        :param save_path: see ``save()``
        :type save_path: ``str``
        """
        data = [self._data_path,
                self.image_shape,
                self.rescale,
                self._shear_range,
                self._zoom_range,
                self._horizontal_flip,
                self._batch_size,
                self.data_classes]

        # Pickle the `data` dictionary.
        save_location = os.path.join(save_path, "{0}_support.p".format(save_name))
        pickle.dump(data, open(save_location, "wb"))

    def save(self, save_name, path=None, overwrite=False):
        """

        Save the weights from a trained model.

        :param save_name: name of the file. Do not include the '.h5' extension as it
                     will be added automatically.
        :type save_name: ``str``
        :param path: path to save the data to. See: ``keras.models.Sequential()``.
        :type path: ``str``
        :param overwrite: overwrite the existing copy of the data
        :type overwrite: ``bool``
        :raises: ``AttributeError`` if ``ImageClassificationCNN().fit()`` is yet to be called.
        """
        self._model_existence_check("saved", "fit", " Alternatively, you can call .load().")
        save_path = self._data_path if (path is None and self._data_path is not None) else path

        # Save the supporting data
        self._support_save_data(save_name, save_path)

        # Save the Model itself
        self.model.save(os.path.join(save_path, "{0}.h5".format(save_name)), overwrite=overwrite)

    def _support_load_data(self, path):
        """

        Loads supporting data saved along with the model

        :param path: see ``load()``.
        :type path: ``str``
        """
        load_location = "{0}_support.p".format(path[:-3])
        with open(load_location, "rb") as f:
            data = pickle.load(f)

        self._data_path = data[0]
        self.image_shape = data[1]
        self.rescale = data[2]
        self._shear_range = data[3]
        self._zoom_range = data[4]
        self._horizontal_flip = data[5]
        self._batch_size = data[6]
        self.data_classes = data[7]

    def load(self, path, override_existing=False, default_model_load=False):
        """

        Load a model from disk.
        
        .. note::
            
            This method expects an additional file ending in "_support.p".
            For example: if ``path='/your/path/my_model.h5'``, a file entitled
            ``'/your/path/my_model_support.p'`` is also expected.

        :param path: path to save the data to.See: ``keras.models.Sequential()``.
        :type path: ``str``
        :param override_existing: If True and a model has already been instantiated, override this replace this model.
                                  Defaults to ``False``.
        :type override_existing: ``bool``
        :param default_model_load: load the default model if ``ImageClassificationCNN().convnet()`` has not been called.
                                   Defaults to ``False``.
        :type default_model_load: ``bool``
        :raises: ``AttributeError`` if a model is currently instantiated.
        """
        if self.model is not None and override_existing is not True:
            raise AttributeError("A model is currently instantiated.\n"
                                 "Set `override_existing` to `True` to replace the existing model.")

        if default_model_load:
            self.convnet()

        self._support_load_data(path)
        self.model = load_model(path)

    def _prediction_labels(self, single_image_prediction):
        """

        Convert a single prediction into a human-readable list of tuples.

        :param single_image_prediction: see ``predict()``
        :type single_image_prediction: list of ndarray arrays.
        :return: a list of tuples where the elements are of the form ``(label, P(label))``
        :rtype: ``list``
        """
        data_classes_reversed = {v: k for k, v in self.data_classes.items()}
        predictions = ((data_classes_reversed[e], i) for e, i in enumerate(single_image_prediction))
        return sorted(predictions, key=lambda x: x[1], reverse=True)

    def predict(self, list_of_images, desc=None, status=True):
        """

        Generate Predictions for a list of images.

        :param list_of_images: a list of paths (strings) to images or ``ndarrays``.
        :type list_of_images: ``list``
        :param desc: description for ``tqdm``.
        :type desc: ``str`` or ``None``
        :param status: True for a tqdm status bar; False for no status bar. Defaults to True.
        :type status: ``bool``
        :return: a list of lists with tuples of the form (name, probability). Defaults to False.
        :rtype: ``list``
        """
        if self.model is None:
            raise AttributeError("Predictions cannot be made until a model is loaded or trained.")

        is_ndarray = [type(i).__name__ == 'ndarray' for i in list_of_images]

        if all(is_ndarray):
            images = list_of_images
        elif any(is_ndarray):
            raise ValueError("Only some of the items in `list_of_images` we found to be `ndarrays`.")
        else:
            images = load_and_scale_images(list_of_images=list_of_images, image_size=self.image_shape,
                                           status=status, grayscale_first=True,
                                           desc='Preparing Images')

        # ToDo: tqdm not working properly (see: https://github.com/bstriner/keras-tqdm).
        return [self._prediction_labels(i) for i in tqdm(self.model.predict(images), desc=desc, disable=not status)]



# coding: utf-8

"""

    Open-i Interface
    ~~~~~~~~~~~~~~~~

"""
import os
import shutil
import pickle
import requests
import numpy as np
import pandas as pd
from math import floor
from copy import deepcopy
from datetime import datetime

from biovida import __version_numeric__

# General Image Support Tools
from biovida.images._image_tools import NoResultsFound, sleep_with_noise

# Database Management
from biovida.images.image_cache_mgmt import (_records_db_merge,
                                             _openi_image_relation_map,
                                             _record_update_dbs_joiner,
                                             _prune_rows_with_deleted_images)

# Interface Support tools
from biovida.images._interface_support.shared import save_records_db

from biovida.images._interface_support.openi.openi_support_tools import (iter_join,
                                                                         url_combine,
                                                                         null_convert,
                                                                         ImageProblemBasedOnText,
                                                                         nonessential_openi_columns)

# Open-i API Parameters Information
from biovida.images._interface_support.openi.openi_parameters import openi_search_information
from biovida.images._interface_support.openi._openi_image_id_processing import image_id_short_gen
from biovida.images._interface_support.openi.openi_text_processing import openi_raw_extract_and_clean

# Cache Management
from biovida.support_tools._cache_management import package_cache_creator

# General Support Tools
from biovida.support_tools.support_tools import (tqdm,
                                                 cln,
                                                 camel_to_snake_case,
                                                 data_frame_col_drop,
                                                 list_to_bulletpoints)


# ----------------------------------------------------------------------------------------------------------
# Searching
# ----------------------------------------------------------------------------------------------------------


class _OpeniSearch(object):
    """

    Searching the Open-i Database.

    """

    def __init__(self):
        self._root_url = 'https://openi.nlm.nih.gov'
        self.current_search = None
        self.search_dict, self.ordered_params = openi_search_information()
        self.search_params = ('image_type', 'rankby', 'article_type', 'subset', 'collection', 'fields',
                              'specialties', 'video', 'exclusions')

    @staticmethod
    def _openi_search_special_case(search_param, blocked, passed):
        """

        Check for invalid search requests when contradictions are possible,
        e.g., `passed` cannot contain `True` and `False`. 

        :param search_param: currently, only `rankby` is valid.
        :param blocked: mutually exclusive (i.e., all these items cannot be passed together).
        :param passed: values actually passed to `search_param`.
        """
        if all(b in passed for b in blocked):
            raise ValueError("`{0}` can only contain one of:\n{1}".format(
                search_param, list_to_bulletpoints(blocked)))

    def _openi_search_check(self, search_arguments):
        """

        Method to check for invalid search requests.

        :param search_arguments: arguments to pass to Open-i's REST API.
        :type search_arguments: ``dict``
        """
        general_error_msg = "'{0}' is not valid for `{1}`.\nValid values for `{1}`:\n{2}"

        if not isinstance(search_arguments['query'], str):
            raise ValueError("`query` must be a string.")

        # Check all other params
        for k, v in search_arguments.items():
            # Halt if query or `v` is NoneType
            if k != 'query' and v is not None:
                # Check type
                if not isinstance(v, (list, tuple)) and v is not None:
                    raise ValueError("Only `lists`, `tuples` or `None` "
                                     "may be passed to `{0}`.".format(k))
                for i in v:
                    if not isinstance(i, str):
                        raise ValueError("`tuples` or `lists` passed to `{0}` "
                                         "must only contain strings.".format(k))
                    # Check if `i` can be converted to a param understood by the Open-i API
                    if i not in self.search_dict[k][1].keys():
                        raise ValueError(general_error_msg.format(
                            i, k, list_to_bulletpoints(self.search_dict[k][1].keys())))

                if k == 'rankby':
                    self._openi_search_special_case(k, blocked=['newest', 'oldest'], passed=v)

    @staticmethod
    def _exclusions_image_type_merge(args, exclusions):
        """
        
        Merge Image type with Exclusions.
        This must be performed because while this module
        separates 'exclusions' it is actually under the 
        'image_type' ('&it') param. in the open-i REST API.
        
        :param args: all arguments passed to ``search()`` to pass to Open-i's REST API.
        :type args: ``dict``
        :param exclusions: a list of exclusions to pass to '&it'.
        :type exclusions: ``list`` or ``None``
        """
        if not isinstance(exclusions, (list, tuple)) and exclusions is not None:
            raise ValueError('`exclusions` must be a `list`, `tuple` or `None`.')

        if exclusions is None or (isinstance(exclusions, (list, tuple)) and not len(exclusions)):
            return args

        if any(e not in ['graphics', 'multipanel'] for e in exclusions):
            raise ValueError("`exclusions` must only include one or all of: 'graphics', 'multipanel'.")

        # Handle handle tuples, then `None`s (with the 'else []').
        args['image_type'] = list(args['image_type']) if isinstance(args['image_type'], (list, tuple)) else []

        # Merge `exclusions` with `image_type`
        args['image_type'] += list(map(lambda x: 'exclude_{0}'.format(x), exclusions))

        return args

    def _search_url_formatter(self, api_search_transform):
        """

        Format ``api_search_transform`` as a complete URL to pass to Open-i.

        :param api_search_transform: a dictionary of parameters and keys in the precise form Open-i expects.
        :type api_search_transform: ``dict``
        :return: a formatted `GET` request.
        :rtype: ``str``
        """
        if 'query' in api_search_transform:
            api_search_transform['query'] = cln(api_search_transform['query']).replace(' ', '+')
        else:
            raise ValueError("No `query` detected.")

        search_term = ""
        for p in self.ordered_params:
            if p in api_search_transform:
                search_term += "{0}{1}={2}".format(("?" if p == 'query' else ""), p, api_search_transform[p])

        return "{0}/retrieve.php{1}".format(self._root_url, search_term)

    @staticmethod
    def _search_probe(search_query, print_results):
        """

        Request results from Open-i.

        :param search_query: a fully format Open-i REST GET command.
        :type search_query: ``str``
        :param print_results: print the number of results found
        :type print_results: ``bool``
        :return: a tuple of the form ``(total number of results, the first result returned)``
        :rtype: ``tuple``
        """
        # Get a sample request
        sample = requests.get(search_query + "&m=1&n=1").json()

        try:
            total = int(float(sample['total']))
        except:
            raise ValueError("Could not obtain total number of results from the Open-i API.")

        if total < 1:
            raise NoResultsFound("\nPlease Try Refining Your Search.")

        if print_results:
            print("\nResults Found: %s." % ('{:,.0f}'.format(total)))

        return total, sample['list'][0]

    def options(self, search_parameter=None, print_options=True):
        """

        Options for parameters of `openi_search()`.

        :param search_parameter: one of: 'image_type', 'rankby', 'article_type', 'subset', 'collection', 'fields',
                                         'specialties', 'video' or `exclusions`. If ``None``, print the parameters
                                         of ``search()``.
        :param print_options: if True, pretty print the options, else return as a ``list``.
        :return: a list of valid values for a given search `search_parameter`.
        :rtype: ``list``
        """
        if search_parameter is not None and not isinstance(search_parameter, str):
            raise TypeError("`search_parameter` must be a string or None.")

        # Terms to blocked from displaying to users if search_parameter != 'exclusions'
        exclusions = ['exclude_graphics', 'exclude_multipanel']

        if search_parameter is None:
            opts = self.search_params
        elif search_parameter == 'exclusions':
            opts = [i.split("_")[1] for i in exclusions]
        else:
            # Get the relevant dict of params
            search_dict_against_param = self.search_dict.get(cln(search_parameter).strip().lower(), None)

            if search_dict_against_param is None:
                raise ValueError("'{0}' is not a valid parameter to pass to the Open-i API.".format(search_parameter))

            # Remove exclusions term
            opts = [i for i in search_dict_against_param[1].keys() if i not in exclusions]
            if not len(opts):
                raise ValueError("Relevant options for '{0}'.".format(search_parameter))

        if print_options:
            print(list_to_bulletpoints(opts))
        else:
            return opts

    @staticmethod
    def _search_clean(k, v):
        """

        Clean the search terms (strings) passed to `OpeniInterface().search()`.

        :param k: Open-i param.
        :type k: ``str``
        :param v: value passed to ``k`` .
        :type v: ``list``
        :return: ``v`` with the strings it contained cleaned. 
        :rtype: ``list``
        """
        return [cln(i).replace(' ', '_').lower() for i in v] if k != 'query' and v is not None else v

    def _api_url_terms(self, k, v):
        """

        Convert values passed into a valid REST GET string
        from a dictionary.

        :param k: Open-i param.
        :type k: ``str``
        :param v: value passed to ``k`` .
        :type v: ``list``
        :return: see description.
        :rtype: ``str``
        """
        return ','.join([self.search_dict[k][1][i] for i in v]) if k != 'query' else v

    def _api_url_param(self, k):
        """
        
        Convert param names into a form the API will understand.
        E.g., 'video' --> '&vid'.
        
        :param k: a human readable term into an actual Open-i param.
        :return: see description.
        :rtype: ``str``
        """
        return self.search_dict[k][0] if k != 'query' else 'query'

    def search(self,
               query=None,
               image_type=None,
               rankby=None,
               article_type=None,
               subset=None,
               collection=None,
               fields=None,
               specialties=None,
               video=None,
               exclusions=None,
               print_results=True):
        """

        Tool to generate a search term (URL) for the NIH's Open-i API.
        The computed term is stored as a class attribute (``INSTANCE.current_search_url``)

        :param query: a search term. ``None`` will be converted to an empty string.
        :type query: ``str`` or ``None``
        :param image_type: see ``OpeniInterface().options('image_type')`` for valid values.
        :type image_type: ``str``, ``list``, ``tuple`` or ``None``
        :param rankby: see ``OpeniInterface().options('rankby')`` for valid values.
        :type rankby: ``str``, ``list``, ``tuple`` or ``None``
        :param article_type: see ``OpeniInterface().options('article_type')`` for valid values.
        :type article_type: ``str``, ``list``, ``tuple`` or ``None``
        :param subset: see ``OpeniInterface().options('subset')`` for valid values.
        :type subset: ``str``, ``list``, ``tuple`` or ``None``
        :param collection: see ``OpeniInterface().options('collection')`` for valid values.
        :type collection: ``str``, ``list``, ``tuple`` or ``None``
        :param fields: see ``OpeniInterface().options('fields')`` for valid values.
        :type fields: ``str``, ``list``, ``tuple`` or ``None``
        :param specialties: see ``OpeniInterface().options('specialties')`` for valid values.
        :type specialties: ``str``, ``list``, ``tuple`` or ``None``
        :param video: see ``OpeniInterface().options('video')`` for valid values. Defaults to ``None``.
        :type video: ``str``, ``list``, ``tuple`` or ``None``
        :param exclusions: one or both of: 'graphics', 'multipanel'. See: ``OpeniInterface.search()``.
        :type exclusions: ``list``, ``tuple`` or ``None``
        :param print_results: if ``True``, print the number of search results.
        :type print_results: ``bool``
        """
        # Remove 'self' and 'print_results' from locals
        args_cleaned = {k: v for k, v in deepcopy(locals()).items()
                        if k not in ['self', 'print_results']}

        # Extract the function arguments and format values.
        args = {k: [v] if isinstance(v, str) and k != 'query' else v
                for k, v in args_cleaned.items() if k != 'exclusions'}

        # Merge `image_type` with `exclusions`
        args = self._exclusions_image_type_merge(args, exclusions)

        # Get the arguments
        search_arguments = {k: '' if k == 'query' and v is None else self._search_clean(k, v)
                            for k, v in args.items()}

        # Save search query
        self.current_search = {k: v for k, v in deepcopy(search_arguments).items() if v is not None}

        # Check if search terms are valid
        self._openi_search_check(search_arguments=search_arguments)

        # Perform transformation
        api_search_transform = {self._api_url_param(k): self._api_url_terms(k, v)
                                for k, v in search_arguments.items() if v is not None}

        # Format `api_search_transform`
        search_url = self._search_url_formatter(api_search_transform=api_search_transform)

        # Unpack the probe containing information on the total number of results and list of results to harvest
        current_search_total, current_search_to_harvest = self._search_probe(search_url, print_results)

        return {"query": args_cleaned,
                "search_url": search_url,
                "current_search_total": current_search_total,
                "current_search_to_harvest": current_search_to_harvest}


# ----------------------------------------------------------------------------------------------------------
# Pull Records from the NIH's Open-i API
# ----------------------------------------------------------------------------------------------------------


class _OpeniRecords(object):
    """

    Obtaining Records from the Open-i Database.

    :param root_url: suggested: 'https://openi.nlm.nih.gov'
    :type root_url: ``str``
    :param date_format: suggested: "%d/%m/%Y" (consider leaving as datetime)
    :type date_format: ``str``
    :param verbose: if ``True`` print additional details.
    :type verbose: ``bool``
    :param cache_path: path to the location of the BioVida cache. If a cache does not exist in this location,
                   one will created. Default to ``None``, which will generate a cache in the home folder.
    :type cache_path: ``str``
    :param req_limit: Defaults to 30 (max allowed by Open-i; see: https://openi.nlm.nih.gov/services.php?it=xg).
    :type req_limit: ``int``
    """

    def __init__(self, root_url, date_format, verbose, cache_path, req_limit=30):
        self.root_url = root_url
        self.date_format = date_format
        self._verbose = verbose
        self._cache_path = cache_path
        self.req_limit = req_limit

        self.records_db = None
        self.download_limit = 100  # set to reasonable default.

        # Sleep Time
        self.records_sleep_time = None

    def openi_bounds(self, total):
        """
        
        Generate a list of tuples which define
        the upper and lower bounds of each request to Open-i
        for records, e.g., ``[(1, 30), ..., (90, total)]``.

        :param total: the total number of results for a given search.
        :type total: ``int``
        :return: see description.
        :rtype: ``list``
        """
        end = 1
        bounds = list()

        if total < 1:
            raise ValueError("'{0}' is an invalid value for total.".format(str(total)))

        if self.download_limit is not None and not isinstance(self.download_limit, int):
            raise ValueError("`download_limit` must be an `int` or `None`.")
        if isinstance(self.download_limit, int) and self.download_limit < 1:
            raise ValueError("`download_limit` cannot be less than 1.")

        if total < self.req_limit:
            return [(1, total)]
        elif self.download_limit is not None and total > self.download_limit:
            download_no = self.download_limit
        else:
            download_no = total

        # Compute the number of steps and floor
        n_steps = int(floor(download_no / self.req_limit))

        for _ in range(n_steps):
            bounds.append((end, end + (self.req_limit - 1)))
            end += self.req_limit

        # Compute the remainder
        remainder = download_no % self.req_limit

        # Add remaining part, if nonzero
        if remainder != 0:
            bounds += [(download_no - remainder + 1, download_no)]

        return bounds, download_no

    @staticmethod
    def openi_bounds_formatter(bounds):
        """

        Format the computed bounds for the Open-i API.

        :param bounds: as returned by ``_OpeniPull().openi_bounds()``
        :type bounds: ``list``
        :return: the inner tuples in ``bounds`` converted to strings for Open-i.
        :rtype: ``list``
        """
        return ["&m={0}&n={1}".format(i[0], i[1]) for i in bounds]

    def date_formatter(self, date_dict):
        """

        Format dates in the 'journal_date' column.

        :param date_dict: a dictionary of the date information
        :type date_dict: ``dict``
        :return: ``date_dict`` as a datetime object.
        :rtype: ``datetime``
        """
        # ToDo: return a datetime object instead of a string.
        if not date_dict:
            return None

        try:
            info = [int(i) if i.isdigit() else None for i in
                    [date_dict['year'], date_dict['month'], date_dict['day']]]
        except:
            return None

        if info[0] is None or (info[1] is None and info[2] is not None):
            return None

        cleaned_info = [1 if i is None or i < 1 else i for i in info]

        # ToDo: add date format guessing -- e.g., ``pd.to_datetime(..., infer_datetime_format=True)``.
        try:
            return datetime(cleaned_info[0], cleaned_info[1], cleaned_info[2]).strftime(self.date_format)
        except:
            return None

    @staticmethod
    def harvest_vector(request_rslt):
        """

        Defines the terms to harvest from the results returned by the Open-i API.
        
        Note:
        
            This method assumes a maximum of one nest in ``request_rslt`` (i.e., ``key: {key: value...}``).

        :param request_rslt: request 'records' from Open-i.
        :type request_rslt: ``dict``
        :return: Convert ``request_rslt`` into a list. Nested dictionaries in ``request_rslt`` are added 
                 the list this method returns as tuples.
        :rtype: ``list``
        """
        time_units = ['day', 'month', 'year']

        to_harvest = list()
        for k, v in request_rslt.items():
            if isinstance(v, str):
                to_harvest.append(k)
            elif isinstance(v, dict) and any(dmy in map(lambda x: x.lower(), v.keys()) for dmy in time_units):
                to_harvest.append(k)
            elif isinstance(v, dict):
                for i in v.keys():
                    to_harvest.append((k, i))

        return to_harvest

    def openi_block_harvest(self, url, bound, to_harvest):
        """

        Download a single block (or 'chunk') of records, e.g, 'records' 1-30.

        :param url: the GET URL for the current search.
        :type url: ``str``
        :param bound: the bound to append to the GET URL, e.g., "&m=1&n=30".
        :type bound: ``str``
        :param to_harvest: yield of ``harvest_vector()``.
        :type to_harvest: ``list``
        :return: a list of dictionaries, where the keys of the dictionaries will become
                 the columns of the ``records_db`` dataframe.
        :rtype: ``list``
        """
        # Request data from the Open-i servers
        req = requests.get(url + bound).json()['list']

        root_url_columns = ('detailed_query_url', 'get_article_figures',
                            'similar_in_collection', 'similar_in_results')

        def append_root_url(item):
            """Check whether or not to add `self.root_url` to a column."""
            return 'img_' in item or any(c == item for c in root_url_columns)

        list_of_dicts = list()
        for item in req:
            # Create an item_dict the dict
            item_dict = dict()
            # Populate current `item_dict`
            for j in to_harvest:
                if isinstance(j, (list, tuple)):
                    item_dict[iter_join(j)] = null_convert(item.get(j[0], {}).get(j[1], None))
                elif j == 'journal_date':
                    item_dict[j] = null_convert(self.date_formatter(item.get(j, None)))
                elif append_root_url(camel_to_snake_case(j)):
                    item_dict[j] = url_combine(self.root_url, null_convert(item.get(j, None)))
                else:
                    item_dict[j] = null_convert(item.get(j, None))

            list_of_dicts.append(item_dict)

        return list_of_dicts

    def _records_pull_engine(self, bounds_list, search_url, to_harvest, records_sleep_time, download_no):
        """
        
        Download all records requested by the user.

        :param bounds_list: a list of bounds (chunks), e.g., ``["&m=1&n=30", "&m=31&n=59"...]``. 
        :type bounds_list: ``list``
        :param search_url: search URL as generated by ``_OpeniSearch().search()``.
        :type search_url: ``str``
        :param to_harvest: yield of ``harvest_vector()``.
        :type to_harvest: ``list``
        :param records_sleep_time: (every x downloads, period of time [seconds])
        :type records_sleep_time: ``tuple`` or ``None``
        :param download_no: the number of records to download.
        :type download_no: ``int``
        :return: a list of dictionaries where the keys are future columns of the ``records_db`` dataframe.
        :rtype: ``list``
        """
        if self._verbose:
            print("\nNumber of Records to Download: {0} (chunk size: {1} records).".format(
                '{:,.0f}'.format(download_no), str(self.req_limit)))

        do_sleep = isinstance(records_sleep_time, (list, tuple)) and len(records_sleep_time) == 2

        harvested_data = list()
        for c, bound in enumerate(tqdm(bounds_list, desc='Obtaining Records', disable=not self._verbose), start=1):
            if do_sleep and c % records_sleep_time[0] == 0:
                sleep_with_noise(amount_of_time=records_sleep_time[1])
            harvested_data += self.openi_block_harvest(search_url, bound=bound, to_harvest=to_harvest)

        return harvested_data

    def records_pull(self,
                     search_url,
                     to_harvest,
                     total,
                     query,
                     pull_time,
                     records_sleep_time,
                     clinical_cases_only,
                     download_limit=None,
                     **kwargs):
        """

        'Walk' along the search query and harvest the data
        in the 'chunk size' given by ``req_limit``.

        :param search_url: search URL as generated by ``_OpeniSearch().search()``.
        :type search_url: ``str``
        :param to_harvest: a list of results to harvest as generated by ``_OpeniSearch().search()``
                          (AKA 'current_search_to_harvest').
        :type to_harvest: ``list``
        :param total: the total number of results (current_search_total) as generated by ``_OpeniSearch().search()``.
        :type total: ``int``
        :param query: a dictionary of parameters and values passed to ``_OpeniSearch().search()``.
        :type query: ``dict``
        :param pull_time: yield of ``datetime.now()`` as as evolved inside ``OpeniInterface.pull()``.
        :type pull_time: ``datetime.datetime``
        :param records_sleep_time: (every x downloads, period of time [seconds])
        :type records_sleep_time: ``tuple`` or ``None``
        :param clinical_cases_only: see ``OpeniInterface().pull()``
        :type clinical_cases_only: ``bool``
        :param download_limit: see ``OpeniInterface().pull()``
        :type download_limit: ``None`` or ``int``
        :return: a complete ``records_db``
        :rtype: ``Pandas DataFrame``
        """
        if isinstance(download_limit, int):
            self.download_limit = download_limit
        elif download_limit is None:
            self.download_limit = None
        else:
            raise TypeError("`download_limit` must be an `int` or `None`")

        # Get a list of lists with the bounds
        bounds, download_no = self.openi_bounds(total)

        # Compute a list of search ranges to pass to the Open-i API
        bounds_list = self.openi_bounds_formatter(bounds)

        harvest = self._records_pull_engine(bounds_list=bounds_list,
                                            search_url=search_url,
                                            to_harvest=self.harvest_vector(to_harvest),
                                            records_sleep_time=records_sleep_time,
                                            download_no=download_no)

        records_db = pd.DataFrame(harvest).fillna(np.NaN)

        if kwargs.get('return_raw', False):
            return records_db

        records_db = openi_raw_extract_and_clean(data_frame=records_db,
                                                 clinical_cases_only=clinical_cases_only,
                                                 verbose=self._verbose,
                                                 cache_path=self._cache_path)

        records_db['query'] = [query] * records_db.shape[0]
        records_db['pull_time'] = [pull_time] * records_db.shape[0]
        records_db['biovida_version'] = [__version_numeric__] * records_db.shape[0]

        self.records_db = records_db
        return self.records_db


# ----------------------------------------------------------------------------------------------------------
# Image Harvesting
# ----------------------------------------------------------------------------------------------------------


class _OpeniImages(object):
    """
    
    Obtaining Images from the Open-i Database.

    :param image_save_location: directory to save the images to.
    :type image_save_location: ``str``
    :param database_save_location: system path to 'openi/databases'.
    :type database_save_location: ``str``
    :param verbose: if ``True`` print additional details.
    :type verbose: ``bool``
    """

    def __init__(self, image_save_location, database_save_location, verbose):
        self.image_save_location = image_save_location
        self._verbose = verbose

        # Database
        self.records_db_images = None
        self.real_time_update_db = None

        self.temp_directory_path = os.path.join(database_save_location, "__temp__")

    def _create_temp_directory_path(self):
        """

        Check if ``self.temp_directory_path`` exists. If not, create it.

        """
        if not os.path.isdir(self.temp_directory_path):
            os.makedirs(self.temp_directory_path)

    def _instantiate_real_time_update_db(self, db_index):
        """

        Create the ``real_time_update_db`` and define the path to the location where it will be saved.

        :param db_index: the index of the ``real_time_update_db`` dataframe (should be from ``records_db``).
        :type db_index: ``Pandas Series``
        """
        real_time_update_columns = ['cached_images_path', 'download_success']

        # Instantiate
        db = pd.DataFrame(columns=real_time_update_columns, index=db_index).replace({np.NaN: None})
        self.real_time_update_db = db

    def _title_image(self, url, image_size):
        """

        Generate a title for the images.

        :param url: the URL to the image.
        :type url: ``str``
        :param image_size: the size of the image, e.g., 'large'.
        :type image_size: ``str``
        :return: see description. 
        :type: ``str``
        """
        # Get the actual file name
        base = os.path.basename(url)

        # Separate the name from the image type
        b_name, image_format = os.path.splitext(base)

        # Generate and clean strings to populate the name format below. Note: 1 = file number
        # (in case medpix has images with multiple segments -- though, it doesn't appear to currently.)
        replacement_terms = map(lambda x: cln(x), (str(1), b_name, image_size, image_format.replace(".", "")))

        # Generate the name for the image
        image_name = "{0}__{1}__{2}.{3}".format(*replacement_terms)

        # Return the save path.
        return os.path.join(self.image_save_location, image_name)

    def _individual_image_harvest(self, index, image_url, image_save_path, block):
        """

        Harvests a single image.

        :param index: the row (index) that it currently being on through.
        :param index: ``int``
        :param image_url: URL to the image.
        :type image_url: ``str``
        :param image_save_path: the location to save the image.
        :type image_save_path: ``str``
        :param block: whether or not to block downloading the image if it does not already exist in the cache.
        :type block: ``bool``
        :return: `1` if an image was downloaded, `0` otherwise.
        :rtype: ``int``
        """
        image_downloaded = 0

        def proceed_with_download(image_save_path):
            return not os.path.isfile(image_save_path)

        try:
            # Only download if the file does not already exist in the cache.
            if proceed_with_download(image_save_path):
                if block:
                    raise ImageProblemBasedOnText

                # Get the image
                page = requests.get(image_url)
                # Save to disk
                with open(image_save_path, 'wb') as image:
                    image.write(page.content)
                image_downloaded = 1

            self.real_time_update_db.set_value(index, 'cached_images_path', image_save_path)
            self.real_time_update_db.set_value(index, 'download_success', True)
        except:
            self.real_time_update_db.set_value(index, 'cached_images_path', np.NaN)
            self.real_time_update_db.set_value(index, 'download_success', False)

        return image_downloaded

    def _pull_images_engine(self, harvesting_information, images_sleep_time, image_size, use_image_caption):
        """

        Use ``_individual_image_harvest()`` to download all of the data (images) in ``harvesting_information``.

        :param harvesting_information: as evolved inside ``harvesting_information``
        :type harvesting_information: ``list``
        :param images_sleep_time: see ``pull_images()``
        :type images_sleep_time: ``tuple`` or ``None``
        :param image_size: see ``pull_images()``
        :type image_size: ``str``
        :param use_image_caption: if ``True`` block downloading of an image if its caption suggests the presence
                                  of problematic image properties (e.g., 'arrows') likely to corrupt
                                  a dataset intended for machine learning. Defaults to ``False``.
        :type use_image_caption: ``bool``
        """
        def block_decision(ipt):
            """Decide whether or not to block the downloading."""
            return use_image_caption == True and isinstance(ipt, (list, tuple)) and len(ipt)

        do_sleep = isinstance(images_sleep_time, (list, tuple)) and len(images_sleep_time) == 2
            
        download_count = 0
        for index, image_url, image_problems_text in tqdm(harvesting_information, desc='Obtaining Images',
                                                          disable=not self._verbose):
            # Generate the save path for the image
            image_save_path = self._title_image(url=image_url, image_size=image_size)

            # Save the image
            download_count += self._individual_image_harvest(index=index,
                                                             image_url=image_url,
                                                             image_save_path=image_save_path,
                                                             block=block_decision(image_problems_text))

            # Sleep when `download_count` 'builds up' to images_sleep_time[0].
            if do_sleep and download_count == images_sleep_time[0]:
                sleep_with_noise(amount_of_time=images_sleep_time[1])
                download_count = 0  # reset

    def pull_images(self,
                    records_db,
                    image_size,
                    images_sleep_time,
                    use_image_caption):
        """

        Pull images based in ``records_db``.

        :param records_db: yield of ``_OpeniRecords().records_pull()``
        :type records_db: ``Pandas DataFrame``
        :param image_size: one of 'grid150', 'large', 'thumb' or 'thumb_large'.
        :type image_size: ``str``
        :param images_sleep_time: tuple of the form: ``(every x downloads, period of time [seconds])``. Defaults to ``(10, 1.5)``.
                                   Note: noise is randomly added to the sleep time by sampling from a normal distribution
                                   (with mean = 0, sd = 0.75).
        :type images_sleep_time: ``tuple`` or ``None``
        :param use_image_caption: if ``True`` block downloading of an image if its caption suggests the presence
                                  of problematic image properties (e.g., 'arrows') likely to corrupt
                                  a dataset intended for machine learning. Defaults to ``False``.
        :type use_image_caption: ``bool``
        :return: `records_db` with the addition of `cached_images_path` and `download_success` columns.
        :rtype: ``Pandas DataFrame``
        """
        self._create_temp_directory_path()
        self.records_db_images = records_db.copy(deep=True)

        settings_path = os.path.join(self.temp_directory_path, "image_pull_settings.p")
        pickle.dump({k: v for k, v in locals().items() if k not in ('self', 'settings_path')},
                    open(settings_path, "wb"))

        self._instantiate_real_time_update_db(db_index=self.records_db_images.index)

        if image_size not in ('grid150', 'large', 'thumb', 'thumb_large'):
            raise ValueError("`image_size` must be one of: 'grid150', 'large', 'thumb' or 'thumb_large'.")
        image_column = "img_{0}".format(image_size)

        # Extract needed information from the `records_db_images` dataframe to loop over.
        harvesting_information = list(zip(*[self.records_db_images.index,
                                            self.records_db_images[image_column],
                                            self.records_db_images['image_problems_from_text']]))

        self._pull_images_engine(harvesting_information=harvesting_information,
                                 images_sleep_time=images_sleep_time,
                                 image_size=image_size,
                                 use_image_caption=use_image_caption)

        return _record_update_dbs_joiner(records_db=self.records_db_images, update_db=self.real_time_update_db)


# ----------------------------------------------------------------------------------------------------------
# Construct Database
# ----------------------------------------------------------------------------------------------------------


class OpeniInterface(object):
    """

    Python Interface for the NIH's `Open-i <https://openi.nlm.nih.gov>`_ API.

    :param cache_path: path to the location of the BioVida cache. If a cache does not exist in this location,
                       one will created. Default to ``None``, which will generate a cache in the home folder.
    :type cache_path: ``str`` or ``None``
    :param verbose: if ``True`` print additional details.
    :type verbose: ``bool``
    """

    def _save_cache_records_db(self):
        """

        Save ``cache_records_db`` to 'disk'.

        """
        self.cache_records_db.to_pickle(self._cache_records_db_save_path)

    def _load_prune_cache_records_db(self, load):
        """

        Load and Prune the ``cache_records_db``.

        :param load: if ``True`` load the ``cache_records_db`` dataframe in from disk.
        :type load: ``bool``
        """
        cache_records_db = pd.read_pickle(self._cache_records_db_save_path) if load else self.cache_records_db
        self.cache_records_db = _prune_rows_with_deleted_images(cache_records_db=cache_records_db,
                                                                columns=['cached_images_path'],
                                                                save_path=self._cache_records_db_save_path)

    def _latent_temp_dir(self):
        """

        Load a '__temp__' folder of image record databases which were not merged (and subsequently destroyed)
        before python exited (the ``pull()`` method, specifically).

        """
        if os.path.isdir(self._Images.temp_directory_path):
            temp_dir = self._Images.temp_directory_path
            latent_pickles = [os.path.join(temp_dir, i) for i in os.listdir(temp_dir) if i.endswith(".p")]
            if len(latent_pickles):
                with open(latent_pickles[0], "rb") as f:
                    settings_dict = pickle.load(f)
                settings_dict_for_pull = {k: v for k, v in settings_dict.items() if k not in ['records_db']}
                settings_dict_for_pull['new_records_pull'] = False  # adding this separately in `pull` breaks in python2

                print("\nResuming Download...")
                self.load_records_db(records_db=settings_dict['records_db'])
                self.pull(**settings_dict_for_pull)

            # `temp_directory_path` will be destroyed when `pull()` exits successfully

    def _openi_cache_records_db_handler(self):
        """

        1. if cache_records_db.p doesn't exist, simply save ``records_db_update`` to disk.
        2. if cache_records_db.p does exist, merge with ``records_db_update`` and then save to disk.

        """
        def rows_to_conserve_func(x):
            return x['download_success'] == True

        def openi_duplicates_handler(data_frame):
            duplicates_subset_columns = ['img_grid150', 'img_large', 'img_thumb',
                                         'img_thumb_large', 'query', 'download_success']

            # Note: Typically biovida_version' and 'pull_time' will be 'in sync'.
            # However, if they are not, this function gives preference to 'biovida_version'
            # s.t. the data harvested with the latest version is given priority.
            data_frame = data_frame.sort_values(by=['biovida_version', 'pull_time'])
            return data_frame.drop_duplicates(subset=duplicates_subset_columns, keep='last')

        if self.cache_records_db is None and self.records_db is None:
            raise ValueError("`current_records_db` and `records_db` cannot both be None.")
        elif self.cache_records_db is not None and self.records_db is None:
            data_frame = self.cache_records_db
            self.cache_records_db = data_frame[
                data_frame.apply(rows_to_conserve_func, axis=1)].reset_index(drop=True)
        elif self.cache_records_db is None and self.records_db is not None:
            data_frame = _openi_image_relation_map(self.records_db)
            self.cache_records_db = data_frame[
                data_frame.apply(rows_to_conserve_func, axis=1)].reset_index(drop=True)
        else:
            self.cache_records_db = _records_db_merge(interface_name='OpeniInterface',
                                                      current_records_db=self.cache_records_db,
                                                      records_db_update=self.records_db,
                                                      columns_with_dicts=('query', 'parsed_abstract'),
                                                      duplicates=openi_duplicates_handler,
                                                      rows_to_conserve_func=rows_to_conserve_func,
                                                      pre_return_func=image_id_short_gen)

        # Save to disk
        self._save_cache_records_db()

        # Destroy the '__temp__' folder
        shutil.rmtree(self._Images.temp_directory_path, ignore_errors=True)

    def __init__(self, cache_path=None, verbose=True):
        self._cache_path = cache_path
        self._verbose = verbose
        self._root_url = 'https://openi.nlm.nih.gov'
        self._date_format = '%d/%m/%Y'

        # Generate Required Caches
        _, self._created_image_dirs = package_cache_creator(sub_dir='images',
                                                            cache_path=cache_path,
                                                            to_create=['openi'],
                                                            nest=[('openi', 'aux'), ('openi', 'raw'),
                                                                  ('openi', 'databases')],
                                                            verbose=verbose,
                                                            requires_medpix_logo=True)

        self._ROOT_PATH = self._created_image_dirs['ROOT_PATH']

        # Instantiate Classes
        self._Search = _OpeniSearch()

        self._Records = _OpeniRecords(root_url=self._root_url,
                                      date_format=self._date_format,
                                      verbose=verbose,
                                      cache_path=cache_path)

        self._Images = _OpeniImages(image_save_location=self._created_image_dirs['raw'],
                                    database_save_location=self._created_image_dirs['databases'],
                                    verbose=verbose)

        self._pull_time = None
        self.current_query = None
        self.current_search_url = None
        self.current_search_total = None
        self._current_search_to_harvest = None

        self.records_db = None

        # Path to cache record db
        self._cache_records_db_save_path = os.path.join(self._created_image_dirs['databases'],
                                                        'openi_cache_records_db.p')

        # Load the cache record database, if it exists
        if os.path.isfile(self._cache_records_db_save_path):
            self._load_prune_cache_records_db(load=True)
        else:
            self.cache_records_db = None

        # Load in a latent database in 'databases/__temp__', if one exists
        self._latent_temp_dir()

    def save_records_db(self, path):
        """

        Save the current ``records_db``.

        :param path: a system path ending with the '.p' file extension.
        :type path: ``str``
        """
        save_records_db(data_frame=self.records_db, path=path)

    def load_records_db(self, records_db):
        """

        Load a ``records_db``

        :param records_db: a system path or ``records_db`` itself.
        :type records_db: ``str`` or ``Pandas DataFrame``
        """
        if isinstance(records_db, pd.DataFrame):
            self.records_db = records_db
        else:
            self.records_db = pd.read_pickle(records_db)
        self._pull_time = self.records_db['pull_time'].iloc[0]
        last_query = self.records_db['query'].iloc[0]
        last_query['print_results'] = False
        self.search(**last_query)

    @property
    def records_db_short(self):
        """Return `records_db` with nonessential columns removed."""
        return data_frame_col_drop(self.records_db, nonessential_openi_columns, 'records_db')

    @property
    def cache_records_db_short(self):
        """Return `cache_records_db` with nonessential columns removed."""
        return data_frame_col_drop(self.cache_records_db, nonessential_openi_columns, 'cache_records_db')

    def options(self, search_parameter=None, print_options=True):
        """

        Options for parameters of ``search()``.

        :param search_parameter: one of: 'image_type', 'rankby', 'article_type', 'subset', 'collection', 'fields',
                                         'specialties', 'video' or `exclusions`. If ``None``, print the parameters
                                         of ``search()``.
        :type search_parameter: ``str``
        :param print_options: if ``True``, pretty print the options, else return as a ``list``. Defaults to ``True``.
        :type print_options: ``bool``
        :return: a list of valid values for a given search ``search_parameter``.
        :rtype: ``list``
        """
        return self._Search.options(search_parameter, print_options)

    def search(self,
               query=None,
               image_type=None,
               rankby=None,
               article_type=None,
               subset=None,
               collection=None,
               fields=None,
               specialties=None,
               video=None,
               exclusions=['graphics'],
               print_results=True):
        """

        Tool to generate a search term (URL) for the NIH's Open-i API.
        The computed term is stored as a class attribute (``INSTANCE.current_search_url``)

        :param query: a search term. ``None`` will be converted to an empty string.
        :type query: ``str`` or ``None``
        :param image_type: see ``OpeniInterface().options('image_type')`` for valid values.
        :type image_type: ``str``, ``list``, ``tuple`` or ``None``
        :param rankby: see ``OpeniInterface().options('rankby')`` for valid values.
        :type rankby: ``str``, ``list``, ``tuple`` or ``None``
        :param article_type: see ``OpeniInterface().options('article_type')`` for valid values. Defaults to 'case_report'.
        :type article_type: ``str``, ``list``, ``tuple`` or ``None``
        :param subset: see ``OpeniInterface().options('subset')`` for valid values.
        :type subset: ``str``, ``list``, ``tuple`` or ``None``
        :param collection: see ``OpeniInterface().options('collection')`` for valid values.
        :type collection: ``str``, ``list``, ``tuple`` or ``None``
        :param fields: see ``OpeniInterface().options('fields')`` for valid values.
        :type fields: ``str``, ``list``, ``tuple`` or ``None``
        :param specialties: see ``OpeniInterface().options('specialties')`` for valid values.
        :type specialties: ``str``, ``list``, ``tuple`` or ``None``
        :param video: see ``OpeniInterface().options('video')`` for valid values.
        :type video: ``str``, ``list``, ``tuple`` or ``None``
        :param exclusions: one or both of: 'graphics', 'multipanel'. Defaults to ``['graphics']``.

                    .. note::

                           Excluding 'multipanel' can result in images that *are* multipanel
                           being returned from Open-i API. For this reason, including 'multipanel'
                           is not currently recommended.

        :type exclusions: ``list``, ``tuple`` or ``None``
        :param print_results: if ``True``, print the number of search results.
        :type print_results: ``bool``

        .. note::

                If passing a single option to ``image_type``, ``rankby``, ``article_type``, ``subset``,
                ``collection``, ``fields``, ``specialties`` or ``video``, a string can be used, e.g.,
                ``...image_type='ct')``. For multiple values, a list or tuple must be used, e.g.,
                ``...image_type=('ct', 'mri')``.

        """
        search = self._Search.search(query=query,
                                     image_type=image_type,
                                     rankby=rankby,
                                     article_type=article_type,
                                     subset=subset,
                                     collection=collection,
                                     fields=fields,
                                     specialties=specialties,
                                     video=video,
                                     exclusions=exclusions,
                                     print_results=print_results)

        self.current_query = search['query']
        self.current_search_url = search['search_url']
        self.current_search_total = search['current_search_total']
        self._current_search_to_harvest = search['current_search_to_harvest']

    def pull(self,
             image_size='large',
             records_sleep_time=(10, 1.5),
             images_sleep_time=(10, 1.5),
             download_limit=100,
             clinical_cases_only=False,
             use_image_caption=False,
             new_records_pull=True):
        """

        Pull (i.e., download) the current search.

        In addition to the columns provided by Open-i, this method will automatically generate the
        following columns by analyzing the pulled data:

        - ``'age'``
        - ``'sex'``
        - ``'ethnicity'``
        - ``'diagnosis'``
        - ``'parsed_abstract'``
        - duration of illness (``'illness_duration_years'``)
        - the imaging modality (e.g., MRI) used, based on the text associated with the image (``'imaging_modality_from_text'``)
        - the plane ('axial', 'coronal' or 'sagittal') of the image (``'image_plane'``)
        - image problems ('arrows', 'asterisks' and 'grids') inferred from the image caption (``'image_problems_from_text'``)

        .. note::

            The 'parsed_abstract' column contains abstracts coerced into dictionaries where the subheadings of the abstract
            form the keys and their associated information form the values. For example, a *MedPix* image will typically yield
            a dictionary with the following keys: 'history', 'finding', 'ddx' (differential diagnosis), 'dxhow' and 'exam'.

        .. warning::

            *MedPix* images include a distinct 'diagnosis' section. For images from other sources, the ``'diagnosis'``
            column is obtained by analyzing the text associated with the image. This analysis could produce inaccuracies.

        :param image_size: one of: 'large', 'grid150', 'thumb', 'thumb_large' or ``None``. Defaults to 'large'.
                          If ``None``, no attempt will be made to download images.

                        .. warning::

                                The analyses performed by the ``image_processing.OpeniImageProcessing`` class are
                                most accurate with large images.

        :type image_size: ``str`` or ``None``
        :param records_sleep_time: tuple of the form: ``(every x downloads, period of time [seconds])``. Defaults to ``(10, 1.5)``.
                                   Note: noise is randomly added to the sleep time by sampling from a normal distribution
                                   (with mean = 0, sd = 0.75).
        :type records_sleep_time: ``tuple`` or ``None``
        :param images_sleep_time: tuple of the form: ``(every x downloads, period of time [seconds])``. Defaults to ``(10, 1.5)``.
                                  Note: noise is randomly added to the sleep time by sampling from a normal distribution
                                  (with mean = 0, sd = 0.75).
        :type images_sleep_time: ``tuple`` or ``None``
        :param download_limit: max. number of results to download. If ``None``, no limit will be imposed
                              (not recommended). Defaults to 100.
        :type download_limit: ``None`` or ``int``
        :param clinical_cases_only: if ``True`` require that the data harvested is of a clinical case. Specifically,
                                    this parameter requires that 'article_type' is one of: 'encounter', 'case_report'.
                                    Defaults to ``False``.

                    .. note::

                        If ``True``, this parameter will often result in fewer records being returned than
                        the ``download_limit``.

        :type clinical_cases_only: ``bool``
        :param use_image_caption: if ``True`` block downloading of an image if its caption suggests the presence
                                  of problematic image properties (e.g., 'arrows') likely to corrupt a dataset.
                                  Defaults to ``False``.
        :type use_image_caption: ``bool``
        :param new_records_pull: if ``True``, download the data for the current search. If ``False``, use ``INSTANCE.records_db``.

            .. note::

               Setting ``new_records_pull=False`` can be useful if one wishes to initially set ``image_size=None``,
               truncate or otherwise modify ``INSTANCE.records_db`` and then download images.

        :type new_records_pull: ``bool``
        :return: a DataFrame with the record information.
                 If ``image_size`` is not None, images will also be harvested and cached.
        :rtype: ``Pandas DataFrame``
        :raises ``ValueError``: if ``search()`` has not been called.
        """
        if not new_records_pull and not isinstance(self.records_db, pd.DataFrame):
            raise TypeError("`records_db` is not a dataframe.")

        if new_records_pull:
            if not isinstance(self.current_search_url, str):
                raise ValueError("`search()` must be called before `pull()`.")
            self._pull_time = datetime.now()
            self.records_db = self._Records.records_pull(search_url=self.current_search_url,
                                                         to_harvest=self._current_search_to_harvest,
                                                         total=self.current_search_total,
                                                         query=self.current_query,
                                                         pull_time=self._pull_time,
                                                         records_sleep_time=records_sleep_time,
                                                         clinical_cases_only=clinical_cases_only,
                                                         download_limit=download_limit)
        elif not isinstance(self.records_db, pd.DataFrame):
            raise TypeError("`records_db` is not a DataFrame.")

        if isinstance(image_size, str):
            self.records_db = self._Images.pull_images(records_db=self.records_db,
                                                       image_size=image_size,
                                                       images_sleep_time=images_sleep_time,
                                                       use_image_caption=use_image_caption)

            # Add the new records_db datafame with the existing `cache_records_db`.
            self._openi_cache_records_db_handler()

        return self.records_db



# coding: utf-8

"""

    Dictionaries of Open-i Params
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""
from biovida.support_tools.support_tools import cln
from biovida.support_tools.support_tools import dict_reverse


def _parser(s):
    """
    
    Tool to convert the lists on https://openi.nlm.nih.gov/services.php?it=xg to dicts.
    
    :param s: a string copied from the URL above
    :type s: ``str``
    :return: a dictionary of params.
    :rtype: ``dict``
    """
    # Split the string into keys and values
    split_string = [i.split('(') for i in cln(s).split(')')]

    # Define a lambda to clean the values
    def lower_join(x):
        return '_'.join(cln(x).split()).strip().lower()

    # Generate the dict
    return {cln(k[1]): lower_join(k[0]) for k in split_string if set(k) != {''}}


openi_video_params = {
    '1': 'true'
}


openi_image_type_params = {
    'c': 'ct',
    'g': 'graphic',
    'm': 'mri',
    'mc': 'microscopy',
    'p': 'pet',
    'ph': 'photograph',
    'u': 'ultrasound',
    'x': 'x_ray',
    'xg': 'exclude_graphics',
    'xm': 'exclude_multipanel'
}


openi_image_type_modality_full = {
    # Written to cohere with yield of `cancer_imaging_parameters.py`'s
    # `CancerImageArchiveParams().dicom_modality_abbreviations()` method.
    'c': 'Computed Tomography (CT)',
    'g': 'Graphic',
    'm': 'Magnetic Resonance Imaging (MRI)',
    'mc': 'Microscopy',
    'p': 'Positron Emission Tomography (PET)',
    'ph': 'Photograph',
    'u': 'Ultrasound',
    'x': 'X-Ray',
}


openi_rankby_params = {
    'r': 'newest',
    'o': 'oldest',
    'd': 'diagnosis',
    'e': 'etiology',
    'oc': 'outcome',
    'pr': 'prevention',
    'pg': 'prognosis',
    't': 'treatment'
}


openi_subset_params = {
    'b': 'basic_science',
    'c': 'clinical_journals',
    'e': 'ethics',
    's': 'systematic_reviews',
    'x': 'chest_x_rays'
}


openi_collection_params = {
    'pmc': 'pubmed',
    'hmd': 'history_of_medicine',
    'iu': 'indiana_u_xray',
    'mpx': 'medpix',
    'usc': 'usc_anatomy',
}


openi_fields_params = {
    'a': 'authors',
    'ab': 'abstracts',
    'c': 'captions',
    'm': 'mentions',
    'msh': 'mesh',
    't': 'titles'
}


openi_article_type_params = {
    'ab': 'abstract',
    'bk': 'book_review',
    'bf': 'brief_report',
    'cr': 'case_report',
    'dp': 'data_paper',
    'di': 'discussion',
    'ed': 'editorial',
    'ib': 'in_brief',
    'in': 'introduction',
    'lt': 'letter',
    'mr': 'meeting_report',
    'ma': 'method_article',
    'ne': 'news',
    'ob': 'obituary',
    'pr': 'product_review',
    'or': 'oration',
    're': 'reply',
    'ra': 'research_article',
    'rw': 'review_article',
    'sr': 'systematic_reviews',
    'rr': 'radiology_report',
    'os': 'orthopedic_slides',
    'hs': 'historical_slide',
    'ot': 'other'
}


openi_specialties_params = {
    'b': 'behavioral_sciences',
    'bc': 'biochemistry',
    'c': 'cancer',
    'ca': 'cardiology',
    'cc': 'critical_care',
    'd': 'dentistry',
    'de': 'dermatology',
    'dt': 'drug_therapy',
    'e': 'emergency_medicine',
    'eh': 'environmental_health',
    'en': 'endocrinology',
    'f': 'family_practice',
    'g': 'gastroenterology',
    'ge': 'genetics',
    'gr': 'geriatrics',
    'gy': 'gynecology_&_obstetrics',
    'h': 'hematology',
    'i': 'immunology',
    'id': 'infectious_diseases',
    'im': 'internal_medicine',
    'n': 'nephrology',
    'ne': 'neurology',
    'nu': 'nursing',
    'o': 'ophthalmology',
    'or': 'orthopedics',
    'ot': 'otolaryngology',
    'p': 'pediatrics',
    'pu': 'pulmonary_diseases',
    'py': 'psychiatry',
    'r': 'rheumatology',
    's': 'surgery',
    't': 'toxicology',
    'u': 'urology',
    'v': 'vascular_diseases',
    'vi': 'virology'
}


openi_api_search_params = {
    'video': ('&vid', openi_video_params),
    'image_type': ('&it', openi_image_type_params),
    'rankby': ('&favor', openi_rankby_params),
    'article_type': ('&at', openi_article_type_params),
    'subset': ('&sub', openi_subset_params),
    'collection': ('&coll', openi_collection_params),
    'fields': ('&fields', openi_fields_params),
    'specialties': ('&sp', openi_specialties_params),
}


def openi_search_information():
    """

    Returns a dictionary of the form:
        ``{search_category: (URL_Parameter, {search term: URL_Parameter}), ...}``

    :return: search information for the Open-i API.
    :rtype: ``dict``
    """
    # Params in the order the Open-i API expects
    ordered_params = ['query', '&it', '&favor', '&at', '&vid', '&sub', '&coll', '&fields', '&sp']

    # Return the openi_api_search_params dict with the dicts nested therein reversed.
    return {k: (v[0], dict_reverse(v[1])) for k, v in openi_api_search_params.items()}, ordered_params



# coding: utf-8

"""

    DisGeNET Interface
    ~~~~~~~~~~~~~~~~~~

"""
import os
import requests
import pandas as pd

# Tool to create required caches
from biovida.support_tools._cache_management import package_cache_creator

# BioVida Support Tools
from biovida.support_tools.support_tools import header, camel_to_snake_case, list_to_bulletpoints

# BioVida Printing Tools
from biovida.support_tools.printing import dict_pprint


# ---------------------------------------------------------------------------------------------
# DisGeNET Reference Data
# ---------------------------------------------------------------------------------------------


_disgenet_delimited_databases = {
    # Source: http://www.disgenet.org/web/DisGeNET/menu/downloads#curated
    # Structure: {database_short_name: {full_name: ..., url: ..., description: ..., number_of_rows_in_header: ...}}
    'all': {
        'full_name': 'All Gene-Disease Associations',
        'url': 'http://www.disgenet.org/ds/DisGeNET/results/all_gene_disease_associations.tsv.gz',
        'description': 'The file contains all gene-disease associations in DisGeNET.',
        'header': 21
    },
    'curated': {
        'full_name': 'Curated Gene-Disease Associations',
        'url': 'http://www.disgenet.org/ds/DisGeNET/results/curated_gene_disease_associations.tsv.gz',
        'description': 'The file contains gene-disease associations from UNIPROT, CTD (human subset), ClinVar, Orphanet,'
                       ' and the GWAS Catalog.',
        'header': 21
    },
    'snp_disgenet': {
        'full_name': 'All SNP-Gene-Disease Associations',
        'url': 'http://www.disgenet.org/ds/DisGeNET/results/all_snps_sentences_pubmeds.tsv.gz',
        'description': 'All SNP-gene-disease associations.',
        'header': 20
    },
}


# ---------------------------------------------------------------------------------------------
# Tools for Harvesting DisGeNET Data
# ---------------------------------------------------------------------------------------------


class DisgenetInterface(object):
    """

    Python Interface for Harvesting Databases from `DisGeNET <http://www.disgenet.org/>`_.

    :param cache_path: location of the BioVida cache. If one does not exist in this location, one will created.
                       Default to ``None`` (which will generate a cache in the home folder).
    :type cache_path: ``str`` or ``None``
    :param verbose: If ``True``, print notice when downloading database. Defaults to ``True``.
    :type verbose: ``bool``
    """

    @staticmethod
    def _disgenet_readme(created_gene_dirs):
        """

        Writes the DisGeNET README to disk.

        :param created_gene_dirs: the dictionary of directories returned by ``_package_cache_creator()``
        :type created_gene_dirs: ``dict``
        """
        save_address = os.path.join(created_gene_dirs['disgenet'], 'DisGeNET_README.txt')

        if not os.path.isfile(save_address):
            readme_url = 'http://www.disgenet.org/ds/DisGeNET/results/readme.txt'
            r = requests.get(readme_url, stream=True)
            with open(save_address, 'wb') as f:
                f.write(r.content)
            header("The DisGeNET README has been downloaded to:\n\n {0}\n\n"
                   "Please take the time to review this document.".format(save_address), flank=False)

    def __init__(self, cache_path=None, verbose=True):
        """

        Initialize the ``DisgenetInterface()`` Class.

        """
        self._verbose = verbose

        # Cache Creation
        ppc = package_cache_creator(sub_dir='genomics',
                                    cache_path=cache_path,
                                    to_create=['disgenet'],
                                    verbose=verbose)
        self.root_path, self._created_gene_dirs = ppc

        # Check if a readme exists.
        self._disgenet_readme(self._created_gene_dirs)

        # Containers for the most recently requested database.
        self.current_database = None
        self.current_database_name = None
        self.current_database_full_name = None
        self.current_database_description = None

    @staticmethod
    def _disgenet_delimited_databases_key_error(database):
        """

        Raises an error when an reference is made to a database not in `_disgenet_delimited_databases.keys()`.

        :param database: `erroneous` database reference.
        :type database: ``str``
        """
        if database not in _disgenet_delimited_databases:
            raise ValueError("'{0}' is an invalid value for `database`.\n`database` must be one of:\n{1}".format(
                str(database), list_to_bulletpoints(_disgenet_delimited_databases.keys())))

    def options(self, database=None, pretty_print=True):
        """

        Disgenet databases which can be downloaded
        as well as additional information about the databases.

        :param database: A database to review. Must be one of: 'all', 'curated', 'snp_disgenet' or ``None``.
                         If a specific database is given, the database's full name and description will be provided.
                         If ``None``, a list of databases which can be downloaded will be returned (or printed).
                         Defaults to ``None``.
        :type database: ``str``
        :param pretty_print: pretty print the information. Defaults to True.
        :type pretty_print: ``bool``
        :return: a ``list`` if `database` is ``None``, else a ``dict`` with the database's full name and description.
        :rtype: ``list`` or ``dict``
        """
        if database is None:
            info = list(_disgenet_delimited_databases.keys())
        elif database in _disgenet_delimited_databases:
            info = {k: v for k, v in _disgenet_delimited_databases[database].items()
                    if k in ['full_name', 'description']}
        else:
            self._disgenet_delimited_databases_key_error(database)

        if pretty_print:
            if database is None:
                print("Available Databases:\n")
                print(list_to_bulletpoints(info))
            else:
                dict_pprint(info)
        else:
            return info

    @staticmethod
    def _df_clean(data_frame):
        """
        
        Clean the dataframe generated by ``pull()``

        :param data_frame:
        :type data_frame: ``Pandas DataFrame``
        :return: see description.
        :rtype: ``Pandas DataFrame``
        """
        # Lower to make easier to match in the future
        data_frame['diseaseName'] = data_frame['diseaseName'].map(
            lambda x: x.lower() if isinstance(x, str) else x, na_action='ignore')

        data_frame.columns = list(map(camel_to_snake_case, data_frame.columns))

        return data_frame

    def pull(self, database, download_override=False):
        """

        Pull (i.e., download) a DisGeNET Database.

        Note: if a database is already cached, it will be used instead of downloading
        (the `download_override` argument can be used override this behaviour).

        :param database: A database to download. Must be one of: 'all', 'curated', 'snp_disgenet' or ``None``.
                         See ``options()`` for more information.
        :type database: ``str``
        :param download_override: If ``True``, override any existing database currently cached and download a new one.
                                  Defaults to ``False``.
        :type download_override: ``bool``
        :return: a DisGeNET database
        :rtype: ``Pandas DataFrame``
        """
        self._disgenet_delimited_databases_key_error(database)
        db_url = _disgenet_delimited_databases[database]['url']
        save_name = "{0}.p".format(db_url.split("/")[-1].split(".")[0])
        save_address = os.path.join(self._created_gene_dirs['disgenet'], save_name)

        if download_override or not os.path.isfile(save_address):
            if self._verbose:
                header("Downloading DisGeNET Database... ", flank=False)
            data_frame = pd.read_csv(db_url,
                                     sep='\t',
                                     header=_disgenet_delimited_databases[database]['header'],
                                     compression='gzip')
            self._df_clean(data_frame).to_pickle(save_address)
        else:
            data_frame = pd.read_pickle(save_address)

        # Cache the database
        self.current_database = data_frame
        self.current_database_name = database
        self.current_database_full_name = _disgenet_delimited_databases[database]['full_name']
        self.current_database_description = _disgenet_delimited_databases[database]['description']

        return data_frame



# coding: utf-8

"""

    Shared Interface Support Tools
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import os
import pandas as pd


def save_records_db(data_frame, path):
    """

    :param data_frame:
    :param path:
    :return:
    """
    if not isinstance(data_frame, pd.DataFrame):
        raise TypeError("`records_db` is not a DataFrame.")
    elif os.path.isdir(path):
        raise IsADirectoryError("'{0}' is a directory.".format(path))
    save_path = "{0}.p".format(path) if not path.endswith(".p") else path
    data_frame.to_pickle(save_path)



# coding: utf-8

"""

    Template Matching
    ~~~~~~~~~~~~~~~~~

"""
import numpy as np
from scipy.misc import imread
from scipy.misc import imresize
from skimage.feature import match_template


# Notes:
#     See: http://scikit-image.org/docs/dev/api/skimage.feature.html#skimage.feature.match_template.
#     Here, this algorithm has been bootstrapped to make it robust against variance in scale.

# ToDo: the solution implemented here, while it works somewhat well, is lacking.
#       The correct approach is to systematically crop the image as long as the signal
#       continues to increases, up to some ceiling value.


def _arange_one_first(start, end, step, precision=1):
    """

    Wrapper for ``numpy.arange()`` where the number '1' is always first.
    Note: zero will be removed, if created, from the final array.

    :param start: the starting value.
    :type start: ``int``
    :param end: the ending value.
    :type end: ``int``
    :param step: the step size.
    :type step: ``int``
    :param precision: number of decimals to evenly round the array.
    :type precision:  ``int``
    :return: an array created by ``numpy.arange()`` where the number `1` is invariably the first element in the array.
    :rtype: ``ndarray``
    """
    arr = np.around(np.arange(start, end, step), precision)
    arr = arr[arr != 0]
    return np.append(1, arr[arr != 1.0])


def _cropper(base, v_prop, h_prop):
    """

    Crops an image horizontally and vertically.

    Notes:

        - increasing ``h_prop`` increases the amount of the image's left side removed.

        - increasing ``v_prop`` increases the amount of the lower part of the image removed.

    :param base: an image represented as a 2D array.
    :type base: ``2D ndarray``
    :param v_prop: the proportion of the image to remove with respect to its height.
    :type v_prop: ``float``
    :param h_prop: the proportion of the image to remove with respect to its width.
    :type h_prop: ``float``
    :return: a cropped image as a 2D array
    :rtype: ``2D ndarray``
    """
    # Note: image.shape = (height, width).

    # Crop to the left
    hcrop_base = base[:, int(base.shape[1] * h_prop):]

    # Crop with respect to height
    return hcrop_base[:int(base.shape[0] * v_prop)]


def _best_guess_location(match_template_result, scaling=1):
    """

    Takes the result of skimage.feature.match_template() and returns (top left x, top left y)
    by selecting the item in ``match_template_result`` with the strongest signal.

    :param match_template_result: the output of from skimage.feature import match_template.
    :type match_template_result: ``ndarray``
    :return: the upper left for location of the strongest peak and the match quality. Form: ``((x, y), match quality)``.
    :rtype: ``tuple``
    """
    x, y = np.unravel_index(np.argmax(match_template_result), match_template_result.shape)[::-1]
    return np.ceil(np.array((x, y)) / float(scaling)), match_template_result.max()


def _robust_match_template_loading(image, param_name):
    """

    Loads images for ``robust_match_template()``.

    :param image: a path to an image or the image as a 2D array
    :type image: ``str`` or ``2D ndarray``
    :param param_name: the name of the parameter which is being loaded (i.e., `pattern_image` or `base_image`.
    :type param_name: ``str``
    :return: an image as an array.
    :rtype: ``2D ndarray``
    """
    if 'ndarray' in str(type(image)):
        return image
    elif isinstance(image, str):
        return imread(image, flatten=True)
    else:
        raise ValueError("`{0}` must either be a `ndarray` or a path to an image.".format(param_name))


def _min_base_rescale(base, pattern, base_resizes, round_to=3):
    """

    Corrects ``base_resizes`` in instances where it would result
    in the ``base`` image being rescaled to a size which is smaller than the ``pattern`` image.

    Notes:

        - if ``abs(base_resizes[1] - base_resizes[0]) < step size`` at the end of the this function,
           only the unit transformation will take place in ``_matching_engine()``.

        - this function cannot handle the rare case where the pattern is larger than the base.

    :param base: the base image as a 2D array.
    :type base: ``2D ndarray``
    :param pattern: the pattern image as a 2D array.
    :type pattern: ``2D ndarray``
    :param base_resizes: the range over which to rescale the base image.Define as a tuple of the
                         form ``(start, end, step size)``.
    :type base_resizes: ``tuple``
    :param round_to: how many places after the decimal to round to. Defaults to 3.
    :type round_to: ``int``
    :return: ``base_resizes`` either 'as is' or updated to prevent the case outlined in this function's description.
    :rtype: ``tuple``
    """
    # ToDo: Does not always block base < pattern.

    # Pick the limiting axis in the base image (smallest)
    smallest_base_axis = min(base.shape)

    # Pick the limiting axis in the base (largest)
    size_floor = max(pattern.shape)

    min_scalar_for_base = float(np.around(size_floor / smallest_base_axis, round_to))
    base_resizes = list(base_resizes)

    # Move the rescale into valid range, if needed
    if base_resizes[0] < min_scalar_for_base:
        base_resizes[0] = min_scalar_for_base
    if base_resizes[1] < min_scalar_for_base:
        base_resizes[1] += min_scalar_for_base

    return tuple(base_resizes)


def _matching_engine(base, pattern, base_resizes, base_image_cropping, end_search_threshold):
    """

    Runs ``skimage.feature.match_template()`` against ``base`` for a given ``pattern``
    at various sizes of the base image.

    :param base: the base image (typically cropped)
    :type base: ``2D ndarray``
    :param pattern: the pattern image
    :type pattern: `2D ndarray``
    :param base_resizes: the range over which to rescale the base image.Define as a tuple of the
                         form ``(start, end, step size)``.
    :type base_resizes: ``tuple``
    :param base_image_cropping: see ``robust_match_template()``.
    :type base_image_cropping: ``tuple``
    :param end_search_threshold: if a match of this quality is found, end the search. Set ``None`` to disable.
    :type end_search_threshold: ``float`` or  ``None``
    :return: a dictionary of matches made by the ``skimage.feature.match_template()`` function
             with the base image scaled by different amounts (represented by the keys).
             The values are ``tuples`` of the form ``(top left corner, bottom right corner, match quality)``.
    :rtype: ``dict``
    """
    # Crop the base image
    base_h_crop = int(base.shape[1] * base_image_cropping[1])
    cropped_base = _cropper(base, v_prop=base_image_cropping[0], h_prop=base_image_cropping[1])

    # Apply tool to ensure the base will always be larger than the pattern
    start, end, step = _min_base_rescale(cropped_base, pattern, base_resizes, round_to=3)

    match_dict = dict()
    for scale in _arange_one_first(start=start, end=end, step=step):
        # Rescale the image
        scaled_cropped_base = imresize(cropped_base, scale, interp='lanczos')

        # ToDo: this try/except should not be needed.
        try:
            template_match_analysis = match_template(image=scaled_cropped_base, template=pattern)

            top_left, match_quality = _best_guess_location(template_match_analysis, scaling=scale)
            top_left_adj = top_left + np.array([base_h_crop, 0])

            bottom_right = top_left_adj + np.floor(np.array(pattern.shape)[::-1] / scale)
            match_dict[scale] = (list(top_left_adj), list(bottom_right), match_quality)

            if isinstance(end_search_threshold, (int, float)):
                if match_quality >= end_search_threshold:
                    break
        except:
            pass

    return match_dict


def _corners_calc(top_left, bottom_right):
    """

    Compute a dict. with a bounding box derived from
    a top left and top right corner

    :param top_left: tuple of the form: (x, y).
    :param top_left: ``tuple``
    :param bottom_right: tuple of the form: (x, y)
    :param bottom_right: ``tuple``
    :return: a dictionary with the following keys: 'top_left', 'top_right', 'bottom_left' and 'bottom_right'.
             Values are keys of the form (x, y).
    :rtype: ``dict``
    """
    d = {'top_left': top_left,
         'top_right': (bottom_right[0], top_left[1]),
         'bottom_left': (top_left[0], bottom_right[1]),
         'bottom_right': bottom_right}
    return {k: tuple(map(int, v)) for k, v in d.items()}


def robust_match_template(pattern_image,
                          base_image,
                          base_resizes=(0.5, 2.5, 0.1),
                          end_search_threshold=0.875,
                          base_image_cropping=(0.15, 0.5)):
    """

    Search for a pattern image in a base image using a algorithm which is robust
    against variation in the size of the pattern in the base image.

    Method: Fast Normalized Cross-Correlation.

    Limitations:

    - Cropping is limited to the the top left of the base image. The can be circumvented by setting
      ``base_image_cropping=(1, 1)`` and cropping ``base_image`` oneself.

    - This function may become unstable in situations where the pattern image is larger than the base image.

    :param pattern_image: the pattern image.

            .. warning::

                    If a `ndarray` is passed to `pattern_image`, it *must* be preprocessed to be a 2D array,
                    e.g., ``scipy.misc.imread(pattern_image, flatten=True)``

    :type pattern_image: ``str`` or ``ndarray``
    :param base_image: the base image in which to look for the ``pattern_image``.

             .. warning::

                    If a `ndarray` is passed to `base_image`, it *must* be preprocessed to be a 2D array,
                    e.g., ``scipy.misc.imread(base_image, flatten=True)``

    :type base_image: ``str`` or ``ndarray``
    :param base_resizes: the range over which to rescale the base image.Define as a tuple of the
                         form ``(start, end, step size)``. Defaults to ``(0.5, 2.0, 0.1)``.
    :type base_resizes: ``tuple``
    :param end_search_threshold: if a match of this quality is found, end the search. Set equal to ``None`` to disable.
                                 Defaults to 0.875.
    :type end_search_threshold: ``float`` or  ``None``
    :param base_image_cropping: the amount of the image to crop with respect to the x and y axis.
                              form: ``(height, width)``. Defaults to ``(0.15, 0.5)``.
                          
    Notes:

    - Decreasing ``height`` will increase the amount of the lower part of the image removed.

    - Increasing ``width`` will increase the amount of the image's left half removed.

    - Cropping more of the base image reduces the probability that the algorithm getting confused.
      However, if the image is cropped too much, the target pattern itself could be removed.

    :type base_image_cropping: ``tuple``
    :return: A dictionary of the form: ``{"bounding_box": ..., "match_quality": ..., "base_image_shape": ...}``.

            - bounding_box (``dict``): ``{'bottom_right': (x, y), 'top_right': (x, y), 'top_left': (x, y), 'bottom_left': (x, y)}``.

            - match_quality (``float``): quality of the match.

            - base_image_shape (``tuple``): the size of the base image provided. Form: ``(width (x), height (y))``.

    :rtype: ``dict``
    """
    pattern = _robust_match_template_loading(pattern_image, "pattern_image")
    base = _robust_match_template_loading(base_image, "base_image")

    match_dict = _matching_engine(base, pattern, base_resizes, base_image_cropping, end_search_threshold)

    if len(list(match_dict.keys())):
        best_match = max(list(match_dict.values()), key=lambda x: x[2])
        bounding_box = _corners_calc(best_match[0], best_match[1])
        match_quality = best_match[2]
    else:
        bounding_box = None
        match_quality = None

    # Return the bounding box, match quality and the size of the base image
    return {"bounding_box": bounding_box, "match_quality": match_quality, "base_image_shape": base.shape[::-1]}


def _box_show(base_image_path, pattern_image_path):
    """

    This function uses matplotlib to show the bounding box for the pattern.

    :param base_image_path: the path to the base image.
    :type base_image_path: ``str``
    :param pattern_image_path: the path to the pattern image.
    :type pattern_image_path: ``str``
    """
    import matplotlib.pyplot as plt
    import matplotlib.patches as patches

    # Load the Images
    base_image = imread(base_image_path, flatten=True)
    pattern_image = imread(pattern_image_path, flatten=True)

    # Run the analysis
    rslt = robust_match_template(pattern_image, base_image)

    # Extract the top left and top right
    top_left = rslt['bounding_box']['top_left']
    bottom_right = rslt['bounding_box']['bottom_right']

    # Compute the width and height
    width = abs(bottom_right[0] - top_left[0])
    height = abs(bottom_right[1] - top_left[1])

    # Show the base image
    fig, (ax1) = plt.subplots(ncols=1, figsize=(5, 5), sharex=True, sharey=True)
    ax1.imshow(base_image, 'gray')

    # Add the bounding box
    ax1.add_patch(patches.Rectangle(top_left, width, height, fill=False, edgecolor="red"))
    fig.show()
